// [[[$article >= 11]]]
/* Copyright (C) 2005  Thomas Petazzoni

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/

#include <sos/assert.h>
#include <sos/types.h>

#include "netdev_generic.h"
#include "net_protos.h"

sos_ret_t sos_eth_recv_packet (struct sos_netdev * dev,
			       struct sos_netdev_pkt *pkt)
{
  struct ethernet * eth_pkt
    = (struct ethernet*) & pkt->contents[pkt->pdu_offset];

  pkt->pdu_offset += sizeof(struct ethernet);
  SOS_ASSERT_FATAL(pkt->pdu_len >= sizeof(struct ethernet));
  pkt->pdu_len    -= sizeof(struct ethernet);

  if (pkt->pdu_len <= 0)
    return SOS_EINVAL;

  switch (ntohs (eth_pkt->type))
    {
    case ETH_TYPE_ARP:
      {
	return sos_arp_recv_packet (dev, pkt);
      }
      break;

    case ETH_TYPE_IPv4:
      {
	return sos_ipv4_recv_packet (dev, pkt);
      }
      break;

    default:
      break;
    }

  return -SOS_ENOSUP;
}


struct sos_netdev_pkt *sos_eth_alloc_pkt (struct sos_netdev * dev,
					  sos_size_t len,
					  sos_bool_t is_for_xmit,
					  sos_ui32_t alloc_flags)
{
  struct sos_netdev_pkt * pkt;
  pkt = sos_netdev_alloc_pkt (dev,
			      len + sizeof(struct ethernet),
			      is_for_xmit,
			      alloc_flags);
  if (NULL == pkt)
    return pkt;

  pkt->pdu_offset += sizeof(struct ethernet);
  pkt->pdu_len   -= sizeof(struct ethernet);

  sos_bochs_printf ("sos_eth_alloc_pkt(): %p (off=%d,len=%d)\n",
		    pkt->contents, pkt->pdu_offset, pkt->pdu_len);

  return pkt;
}


sos_ret_t sos_eth_send_packet (struct sos_netdev *dev,
			       struct sos_netdev_pkt *pkt,
			       const sos_ui8_t *dstmac,
			       sos_ui16_t type)
{
  struct ethernet *eth_pkt;

  SOS_ASSERT_FATAL (pkt->pdu_offset >= sizeof (struct ethernet));
  pkt->pdu_offset -= sizeof (struct ethernet);
  pkt->pdu_len    += sizeof (struct ethernet);
  SOS_ASSERT_FATAL (pkt->pdu_len <= pkt->contents_size);

  eth_pkt = (struct ethernet *) & pkt->contents[pkt->pdu_offset];

  sos_bochs_printf ("sos_eth_send_packet() %p %p (off=%d,len=%d)\n",
		    pkt->contents, eth_pkt, pkt->pdu_offset, pkt->pdu_len);

  memcpy (eth_pkt->src_mac_addr, dev->hwaddr, sizeof (eth_pkt->src_mac_addr));
  memcpy (eth_pkt->dest_mac_addr, dstmac, sizeof (eth_pkt->dest_mac_addr));
  eth_pkt->type = htons(type);

  return sos_netdev_send_pkt(dev, pkt);
}

// [[[]]]
