// [[[$article >= 11]]]
/* Copyright (C) 2005  Thomas Petazzoni

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/

#include <sos/types.h>
#include <sos/klibc.h>

#include "netdev_generic.h"
#include "net_protos.h"

sos_ret_t sos_icmpv4_recv_packet (struct sos_netdev * dev,
				  struct ip * req_ip,
				  struct sos_netdev_pkt * req_pkt)
{
  struct sos_netdev_pkt *reply_pkt;
  struct icmp    * req_icmp, * reply_icmp;

  if (req_pkt->pdu_len < sizeof(struct icmp))
    return -SOS_EINVAL;

  req_icmp = (struct icmp *) & req_pkt->contents[req_pkt->pdu_offset];

  /* Ignore non-request ICMP */
  if (req_icmp->type != ICMP_ECHO_REQUEST)
    return SOS_OK;

  reply_pkt = sos_ipv4_alloc_pkt (dev, req_pkt->pdu_len, TRUE, 0);
  if (NULL == reply_pkt)
    return -SOS_ENOMEM;

  reply_icmp = (struct icmp *)
    & reply_pkt->contents[reply_pkt->pdu_offset];

  memcpy (reply_icmp, req_icmp, req_pkt->pdu_len);

  /* Exchange source/destination */
  reply_icmp->type = ICMP_ECHO_REPLY;
  reply_icmp->checksum = 0;
  reply_icmp->checksum = ipv4_checksum(reply_icmp, reply_pkt->pdu_len);

  sos_bochs_printf ("sos_icmp_send_packet() %p %p (off=%d,len=%d)\n",
		    reply_pkt->contents, reply_icmp,
		    reply_pkt->pdu_offset, reply_pkt->pdu_len);

  return sos_ipv4_send_packet (dev, reply_pkt,
			       ntohl(req_ip->dest_addr), ntohl(req_ip->src_addr),
			       IP_PROTO_ICMP);
}

// [[[]]]
