// [[[$article >= 11]]]
/* Copyright (C) 2005  Thomas Petazzoni

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/

/* UDP Header, defined by RFC 768,
   http://www.ietf.org/rfc/rfc768.txt. Header length is 8 bytes. */
struct udp {
  sos_ui16_t src_port;
  sos_ui16_t dst_port;
  sos_ui16_t length;
  sos_ui16_t checksum;
  char payload[0];
} __attribute__((packed));

/* ICMP Header, defined by RFC 792,
   http://www.faqs.org/rfcs/rfc792.html. Header length is 6 bytes. */
struct icmp {
#define ICMP_ECHO_REPLY   0x0
#define ICMP_ECHO_REQUEST 0x8
  sos_ui8_t type;
  sos_ui8_t code;
  sos_ui16_t checksum;
  sos_ui16_t sequence;
  char payload[0];
} __attribute__((packed));

/* IPv4 Header, defined by RFC 791,
   http://www.ietf.org/rfc/rfc791.txt. Header length is 20 bytes */
struct ip {
  /* header_length and version are in this order because x86 is a
     little endian architecture. */
  sos_ui8_t header_length:4;
  sos_ui8_t version:4;
  sos_ui8_t services;
  sos_ui16_t length;

  sos_ui16_t identification;
  sos_ui16_t fragment;

  sos_ui8_t ttl;

#define IP_PROTO_ICMP  1
#define IP_PROTO_TCP   6
#define IP_PROTO_UDP  17
  sos_ui8_t protocol;
  sos_ui16_t checksum;
  sos_ui32_t src_addr;
  sos_ui32_t dest_addr;
  char payload[0];
} __attribute__((packed));

/* 14 bytes */
struct ethernet {
  sos_ui8_t  dest_mac_addr[6];
  sos_ui8_t  src_mac_addr[6];

  sos_ui16_t type;
#define ETH_TYPE_IPv4 0x0800
#define ETH_TYPE_ARP  0x0806

  char payload[0];
} __attribute__((packed));

#define NETADDR_2_IPV4(netaddr) ((netaddr[0] << 24) | (netaddr[1] << 16) | \
                                 (netaddr[2] << 8) | netaddr[3])

sos_ret_t sos_icmpv4_recv_packet (struct sos_netdev * dev,
				  struct ip * req_ip,
				  struct sos_netdev_pkt * req_pkt);

struct sos_netdev_pkt * sos_ipv4_alloc_pkt (struct sos_netdev *dev,
					    sos_size_t len,
					    sos_bool_t is_for_xmit,
					    sos_ui32_t flags);

sos_ui16_t ipv4_checksum (const void *buf, unsigned int hdr_len);

sos_ret_t sos_ipv4_send_packet (struct sos_netdev * dev,
				struct sos_netdev_pkt * pkt,
				sos_ui32_t src_addr, sos_ui32_t dest_addr,
				sos_ui8_t protocol);

sos_ret_t sos_arp_send_packet (struct sos_netdev * dev,
			       struct sos_netdev_pkt *pkt,
			       sos_ui32_t dst_ip,
			       sos_ui16_t eth_type);

sos_ret_t sos_udp_recv_packet (struct sos_netdev * dev,
			       struct ip * req_ip,
			       struct sos_netdev_pkt * req_pkt);

sos_ret_t sos_udp_send_packet (struct sos_netdev * dev,
			       struct sos_netdev_pkt * pkt,
			       sos_ui16_t src_port, sos_ui16_t dst_port,
			       sos_ui32_t src_addr, sos_ui32_t dst_addr);

struct sos_netdev_pkt * sos_udp_alloc_pkt (struct sos_netdev *dev,
					   sos_size_t len,
					   sos_bool_t is_for_xmit,
					   sos_ui32_t flags);

// [[[]]]
