// [[[$article >= 11]]]
/* Copyright (C) 2005      David Decotigny

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/
#ifndef _SOS_PCI_H_
#define _SOS_PCI_H_


/** Basic description of a PCI 32bits resource (memory-mapped range or
    i/o mapped range) */
struct sos_pci_resource
{
  enum { SOS_PCI_RESOURCE_MEM, SOS_PCI_RESOURCE_IO } type;
  sos_ui32_t base_address;
  sos_ui32_t length;
};


/** Very basic description of a PCI device */
struct sos_pci_device
{
  sos_ui8_t  bus_id;
  sos_ui8_t  device_id; /**< identifier of the device on the bus */

  /** True when a driver controls this device */
  sos_bool_t is_handled;

  /** Indicative fields to help PCI bus scanning */
  sos_ui32_t vendor_id;
  sos_ui16_t product_id;
  sos_ui32_t class_code;
  sos_ui8_t  revision;

  sos_ui8_t  irq_line;

  /* Each device can have up to 6 BAR */
  struct sos_pci_resource resource[6];

  struct sos_pci_device *prev, *next;
};


sos_ret_t sos_pci_subsystem_setup();


/**
 * Function called for the devices on the PCI buses matching
 * filter_vendor_id/filter_product_id. A PCI_FILTER_MATCH_ALL value
 * means that the function should be called whatever the real value of
 * the field
 *
 * @note this function is responsible to set the is_handled bit if the
 * driver handles the device
 *
 * @return FALSE when the scanning should be stopped
 *
 * @note the "dev" descriptor returned is guaranteed to be persistent
 * for the whole life of the system
 */
typedef sos_bool_t
  (*sos_pci_scan_function_t)(struct sos_pci_device * dev,
			     void *custom_data);

#define SOS_PCI_ID_ANY 0xffffffff

sos_ret_t sos_pci_scan_devices(sos_ui32_t filter_vendor_id,
			       sos_ui32_t filter_product_id,
			       sos_pci_scan_function_t scan_fct,
			       sos_bool_t only_unhandled_devices,
			       void *custom_data);

#endif /* _SOS_PCI_H_ */
// [[[]]]
