// [[[$article >= 11]]]
/* Copyright (C) 2005  Thomas Petazzoni

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/

#include <sos/types.h>

#include "netdev_generic.h"
#include "net_protos.h"

struct sos_netdev_pkt * sos_udp_alloc_pkt (struct sos_netdev *dev,
					   sos_size_t len,
					   sos_bool_t is_for_xmit,
					   sos_ui32_t flags)
{
  struct sos_netdev_pkt *pkt;

  pkt = sos_ipv4_alloc_pkt (dev, len + sizeof (struct udp),
			    is_for_xmit, flags);
  if (NULL == pkt)
    return NULL;

  pkt->pdu_offset += sizeof (struct udp);
  pkt->pdu_len    -= sizeof (struct udp);

  sos_bochs_printf ("[UDP] sos_udp_alloc_pkt(): %p (off=%d,len=%d)\n",
		    pkt->contents, pkt->pdu_offset, pkt->pdu_len);

  return pkt;
}

sos_ret_t sos_udp_recv_packet (struct sos_netdev * dev,
			       struct ip * req_ip,
			       struct sos_netdev_pkt * req_pkt)
{
  struct udp *udp;

  udp = (struct udp *) & req_pkt->contents [req_pkt->pdu_offset];

  sos_bochs_printf ("Received UDP packet. src_port=%d, dst_port=%d, length=%d\n",
		    ntohs(udp->src_port), ntohs(udp->dst_port), ntohs(udp->length));
  sos_bochs_printf ("Contents: %s\n",
		    udp->payload);

  return SOS_OK;
}

sos_ret_t sos_udp_send_packet (struct sos_netdev * dev,
			       struct sos_netdev_pkt * pkt,
			       sos_ui16_t src_port, sos_ui16_t dst_port,
			       sos_ui32_t src_addr, sos_ui32_t dst_addr)
{
  struct udp *udp;
  struct udp_checksum_pseudo_header {
    sos_ui32_t src_addr;
    sos_ui32_t dst_addr;
    sos_ui8_t  zero;
    sos_ui8_t  protocol;
    sos_ui16_t length;
  } __attribute__((packed));
  struct udp_checksum_pseudo_header *ucph;

  sos_bochs_printf ("[UDP] Contents sz=%d\n", pkt->pdu_len);

  pkt->pdu_offset -= sizeof(struct udp);
  pkt->pdu_len    += sizeof(struct udp);

  udp = (struct udp *) & pkt->contents [pkt->pdu_offset];

  udp->src_port = htons (src_port);
  udp->dst_port = htons (dst_port);
  udp->length   = htons (pkt->pdu_len);
  udp->checksum = 0;

  ucph = (struct udp_checksum_pseudo_header *)
    & pkt->contents [pkt->pdu_offset - sizeof (struct udp_checksum_pseudo_header)];

  ucph->src_addr = htonl (src_addr);
  ucph->dst_addr = htonl (dst_addr);
  ucph->protocol = IP_PROTO_UDP;
  ucph->length   = htons(pkt->pdu_len);
  ucph->zero     = 0;

  sos_bochs_printf ("[UDP] Checksum length %d\n",
		    pkt->pdu_len + sizeof(struct udp_checksum_pseudo_header));

  udp->checksum  =
    ipv4_checksum (ucph, pkt->pdu_len + sizeof(struct udp_checksum_pseudo_header));

  return sos_ipv4_send_packet (dev, pkt, src_addr, dst_addr, IP_PROTO_UDP);
}

// [[[]]]
