/* Copyright (C) 2004  The SOS Team

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA. 
*/

#include <sos/errno.h>

/* Include definitions of the multiboot standard */
#include <bootstrap/multiboot.h>
// [[[$article >= 2]]]
#include <hwcore/idt.h>
#include <hwcore/gdt.h>
#include <hwcore/irq.h>
#include <hwcore/exception.h>
#include <hwcore/i8254.h>
// [[[]]]
// [[[$article >= 3]]]
#include <sos/list.h>
#include <sos/physmem.h>
// [[[]]]
// [[[$article >= 4]]]
#include <hwcore/paging.h>
// [[[]]]
// [[[$article >= 7]]]
#include <hwcore/mm_context.h>
#include <hwcore/swintr.h>
// [[[]]]
// [[[$article == 4]]]
#include <sos/list.h>
// [[[]]]
// [[[$article >= 5]]]
#include <sos/kmem_vmm.h>
#include <sos/kmalloc.h>
// [[[]]]
// [[[$article >= 6.5]]]
#include <sos/time.h>
#include <sos/thread.h>
// [[[]]]
// [[[$article >= 7]]]
#include <sos/process.h>
// [[[]]]
// [[[$article >= 7.5]]]
#include <sos/umem_vmm.h>
// [[[]]]
#include <sos/klibc.h>
#include <sos/assert.h>
#include <drivers/x86_videomem.h>
#include <drivers/bochs.h>
// [[[$article >= 7]]]
#include <sos/calcload.h>
// [[[]]]
// [[[$article >= 7.5]]]
#include <sos/umem_vmm.h>
#include <sos/binfmt_elf32.h>
#include <drivers/zero.h>
// [[[]]]
// [[[$article >= 7.5]]]
#include <sos/fs.h>
#include <drivers/fs_virtfs.h>
// [[[]]]
// [[[$article >= 9]]]
#include <drivers/devices.h>
#include <drivers/mem.h>
#include <drivers/tty.h>
#include <drivers/console.h>
// [[[]]]
// [[[$master || $article >= 9]]]
#include <drivers/serial.h>
// [[[]]]
// [[[$article >= 9.5]]]
#include <sos/blkdev.h>
#include <sos/fs_pagecache.h>
#include <drivers/ide.h>
// [[[]]]
// [[[$article >= 10]]]
#include <drivers/fs_ext2.h>
// [[[]]]
// [[[$article >= 11]]]
#include <drivers/pci.h>
#include <sos/socket.h>
// [[[]]]

// [[[$master]]]
void _trc_print_exit(void *s)
{ sos_bochs_printf("[2m</EXIT(%s)>\n[m", *(char**)s); }
// [[[]]]
// [[[$master && $article < 6]]]
void guard_stack(char *str)
{
#if 0
  sos_vaddr_t here = (sos_vaddr_t)&here;
  sos_vaddr_t stack_lowermark = bootstrap_stack_bottom + 1024;
  if ((here <= stack_lowermark) || (here >= bootstrap_stack_top))
  {
    sos_bochs_printf("[%s] Current stack=%x is %s [%x, %x[ (real bottom=%x)\n",
		     str?str:"???",
		     here, (here <= stack_lowermark)?"below":"above",
		     stack_lowermark,
		     bootstrap_stack_top, bootstrap_stack_bottom);

    int nested_level;
    struct bt_data btd;

    bt_init(& btd, NULL);
    for (nested_level = 0 ;
         bt_next(& btd)
	   /* && (nested_level <= 30) */
	   && ((sos_vaddr_t)btd.frame <= bootstrap_stack_top) ;
         nested_level ++)
      {
	unsigned p0=0, p1=0, p2=0, p3=0;
	
	/* Avoids reading above the stack (or else: will get a #PF...) */
	if (bootstrap_stack_top - btd.params_array > 12)
	  p3 = (unsigned) *((unsigned *)(btd.params_array+12));
	if (bootstrap_stack_top - btd.params_array > 8)
	  p2 = (unsigned) *((unsigned *)(btd.params_array+8));
	if (bootstrap_stack_top - btd.params_array > 4)
	  p1 = (unsigned) *((unsigned *)(btd.params_array+4));
	if (bootstrap_stack_top - btd.params_array > 0)
	  p0 = (unsigned) *((unsigned *)(btd.params_array+0));
      sos_bochs_printf("bt[%d] = ret eip 0x%x (0x%x, 0x%x, 0x%x, 0x%x)\n",
		       nested_level, btd.next_instruction, p0, p1, p2, p3);

      }

    SOS_ASSERT_FATAL(! "Stack overflow");
  }

#endif
}

// [[[]]]
// [[[$article >= 2]]]
/* Helper function to display each bits of a 32bits integer on the
   screen as dark or light carrets */
void display_bits(unsigned char row, unsigned char col,
		  unsigned char attribute,
		  sos_ui32_t integer)
{
  int i;
  /* Scan each bit of the integer, MSb first */
  for (i = 31 ; i >= 0 ; i--)
    {
      /* Test if bit i of 'integer' is set */
      int bit_i = (integer & (1 << i));
      /* Ascii 219 => dark carret, Ascii 177 => light carret */
      unsigned char ascii_code = bit_i?219:177;
      sos_x86_videomem_putchar(row, col++,
			       attribute,
			       ascii_code);
    }
}

// [[[]]]
// [[[$article >= 2]]]
/* Clock IRQ handler */
static void clk_it(int intid)
{
  static sos_ui32_t clock_count = 0;

  display_bits(0, 48,
	       SOS_X86_VIDEO_FG_LTGREEN | SOS_X86_VIDEO_BG_BLUE,
	       clock_count);
  clock_count++;
// [[[$article >= 6.5]]]

  /* Execute the expired timeout actions (if any) */
  sos_time_do_tick();
// [[[]]]
// [[[$article >= 7]]]

  /* Update scheduler statistics and status */
  sos_sched_do_timer_tick();
// [[[]]]
}
// [[[]]]
// [[[$article >= 6]]]


/* ======================================================================
 * Page fault exception handling
 */

// [[[$enclosing && $article < 6.5]]]
/* Helper function to dump a backtrace on bochs and/or the console */
static void dump_backtrace(const struct sos_cpu_state *cpu_state,
			   sos_vaddr_t stack_bottom,
			   sos_size_t  stack_size,
			   sos_bool_t on_console,
			   sos_bool_t on_bochs)
{
  static void backtracer(sos_vaddr_t PC,
			 sos_vaddr_t params,
			 sos_ui32_t depth,
			 void *custom_arg)
    {
      sos_ui32_t invalid = 0xffffffff, *arg1, *arg2, *arg3, *arg4;

      /* Get the address of the first 3 arguments from the
	 frame. Among these arguments, 0, 1, 2, 3 arguments might be
	 meaningful (depending on how many arguments the function may
	 take). */
      arg1 = (sos_ui32_t*)params;
      arg2 = (sos_ui32_t*)(params+4);
      arg3 = (sos_ui32_t*)(params+8);
      arg4 = (sos_ui32_t*)(params+12);

      /* Make sure the addresses of these arguments fit inside the
	 stack boundaries */
#define INTERVAL_OK(b,v,u) ( ((b) <= (sos_vaddr_t)(v)) \
                             && ((sos_vaddr_t)(v) < (u)) )
      if (!INTERVAL_OK(stack_bottom, arg1, stack_bottom + stack_size))
	arg1 = &invalid;
      if (!INTERVAL_OK(stack_bottom, arg2, stack_bottom + stack_size))
	arg2 = &invalid;
      if (!INTERVAL_OK(stack_bottom, arg3, stack_bottom + stack_size))
	arg3 = &invalid;
      if (!INTERVAL_OK(stack_bottom, arg4, stack_bottom + stack_size))
	arg4 = &invalid;

      /* Print the function context for this frame */
      if (on_bochs)
	sos_bochs_printf("[%d] PC=0x%x arg1=0x%x arg2=0x%x arg3=0x%x\n",
			 (unsigned)depth, (unsigned)PC,
			 (unsigned)*arg1, (unsigned)*arg2,
			 (unsigned)*arg3);

      if (on_console)
	sos_x86_videomem_printf(23-depth, 3,
				SOS_X86_VIDEO_BG_BLUE
				  | SOS_X86_VIDEO_FG_LTGREEN,
				"[%d] PC=0x%x arg1=0x%x arg2=0x%x arg3=0x%x arg4=0x%x",
				(unsigned)depth, PC,
				(unsigned)*arg1, (unsigned)*arg2,
				(unsigned)*arg3, (unsigned)*arg4);
      
    }

  sos_backtrace(cpu_state, 15, stack_bottom, stack_size, backtracer, NULL);
}

// [[[]]]

/* Page fault exception handler with demand paging for the kernel */
static void pgflt_ex(int intid, struct sos_cpu_state *ctxt)
{
  static sos_ui32_t demand_paging_count = 0;
// [[[$article >=7.5]]]
  struct sos_thread * cur_thr = sos_thread_get_current();
// [[[]]]
  sos_vaddr_t faulting_vaddr  = sos_cpu_context_get_EX_faulting_vaddr(ctxt);
  sos_paddr_t ppage_paddr;

// [[[$master && $article >=7.5]]]

/*   sos_bochs_printf("#PF %x\n", faulting_vaddr); */
/*   if (sos_thread_get_current()->process) */
/*     sos_dump_as(sos_process_get_address_space(sos_thread_get_current()->process), "PF"); */
// [[[]]]
// [[[$article >= 7]]]
// [[[$enclosing && $article < 7.5]]]
  if (sos_cpu_context_is_in_user_mode(ctxt))
// [[[]]]
// [[[$enclosing && $article >= 7.5]]]
  if (sos_cpu_context_is_in_user_mode(ctxt)
      || (cur_thr->fixup_uaccess.return_vaddr))
// [[[]]]
    {
      // [[[$enclosing && $article >= 7.5]]]
      __label__ unforce_address_space;
      sos_bool_t need_to_prepare_user_space_access;
      sos_ui32_t errcode = sos_cpu_context_get_EX_info(ctxt);

      /* Do we already try to access an address space ? */
      need_to_prepare_user_space_access
	= (sos_thread_get_current()->squatted_address_space == NULL);

      if (need_to_prepare_user_space_access)
        /* No: Need to configure the interrupted thread user space */
	sos_thread_prepare_user_space_access(NULL, 0);

      if (SOS_OK ==
	  sos_umem_vmm_try_resolve_page_fault(faulting_vaddr,
					      errcode & (1 << 1),
					      TRUE))
	goto unforce_address_space;

      /* If the page fault occured in kernel mode, return to kernel to
	 the fixup address */
      if (! sos_cpu_context_is_in_user_mode(ctxt))
	{
	  cur_thr->fixup_uaccess.faulted_uaddr = faulting_vaddr;
	  sos_cpu_context_set_EX_return_address(ctxt,
						cur_thr->fixup_uaccess.return_vaddr);
	  goto unforce_address_space;
	}

      // [[[]]]
      // [[[$enclosing && $article < 7.5]]]
      /* User-mode page faults are considered unresolved for the
	 moment */
      // [[[]]]
// [[[$enclosing && $article >= 7.5]]]
      if (need_to_prepare_user_space_access)
	sos_thread_end_user_space_access();

// [[[]]]
      sos_bochs_printf("\e[35mTHR %p: Unresolved USER page Fault at instruction 0x%x on access to address 0x%x (info=%x)!\e[m\n",
		       (void*)sos_thread_get_current(),
		       sos_cpu_context_get_PC(ctxt),
		       (unsigned)faulting_vaddr,
		       (unsigned)sos_cpu_context_get_EX_info(ctxt));
      sos_bochs_printf("Terminating User thread\n");
      sos_thread_exit();
// [[[$enclosing && $article >= 7.5]]]

    unforce_address_space:
      if (need_to_prepare_user_space_access)
	sos_thread_end_user_space_access();
      return;
// [[[]]]
    }

// [[[]]]
  /* Check if address is covered by any VMM range */
  if (! sos_kmem_vmm_is_valid_vaddr(faulting_vaddr))
    {
      /* No: The page fault is out of any kernel virtual region. For
	 the moment, we don't handle this. */
// [[[$enclosing && $article < 6.5]]]
      dump_backtrace(ctxt,
		     bootstrap_stack_bottom,
		     bootstrap_stack_size,
		     TRUE, TRUE);
// [[[]]]
      sos_display_fatal_error("Unresolved page Fault at instruction 0x%x on access to address 0x%x (info=%x)!",
			      sos_cpu_context_get_PC(ctxt),
			      (unsigned)faulting_vaddr,
			      (unsigned)sos_cpu_context_get_EX_info(ctxt));
      SOS_ASSERT_FATAL(! "Got page fault (note: demand paging is disabled)");
    }


  /*
   * Demand paging in kernel space
   */
 
  /* Update the number of demand paging requests handled */
  demand_paging_count ++;
  display_bits(0, 0,
	       SOS_X86_VIDEO_FG_LTRED | SOS_X86_VIDEO_BG_BLUE,
	       demand_paging_count);

  /* Allocate a new page for the virtual address */
  ppage_paddr = sos_physmem_ref_physpage_new(FALSE);
  if (! ppage_paddr)
    SOS_ASSERT_FATAL(! "TODO: implement swap. (Out of mem in demand paging because no swap for kernel yet !)");
  SOS_ASSERT_FATAL(SOS_OK == sos_paging_map(ppage_paddr,
					    SOS_PAGE_ALIGN_INF(faulting_vaddr),
					    FALSE,
					    SOS_VM_MAP_PROT_READ
					    | SOS_VM_MAP_PROT_WRITE
					    | SOS_VM_MAP_ATOMIC));
  sos_physmem_unref_physpage(ppage_paddr);

  /* Ok, we can now return to interrupted context */
}

// [[[]]]
// [[[$article == 2]]]

/* Division by zero exception handler */
static void divide_ex(int exid)
{
  static sos_ui32_t div_count = 0;
  display_bits(0, 0,
	       SOS_X86_VIDEO_FG_LTRED | SOS_X86_VIDEO_BG_BLUE,
	       div_count);
  div_count++;
}
// [[[]]]
// [[[$article == 3]]]

#define MY_PPAGE_NUM_INT 511
struct my_ppage
{
  sos_ui32_t before[MY_PPAGE_NUM_INT];
  struct my_ppage *prev, *next;
  sos_ui32_t after[MY_PPAGE_NUM_INT];
}; /* sizeof() Must be <= 4kB */

static void test_physmem()
{
  /* We place the pages we did allocate here */
  struct my_ppage *ppage_list, *my_ppage;
  sos_count_t num_alloc_ppages = 0, num_free_ppages = 0;

  ppage_list = NULL;
  while ((my_ppage = (struct my_ppage*)sos_physmem_ref_physpage_new(FALSE))
	 != NULL)
    {
      int i;
      num_alloc_ppages++;
      
      /* Print the allocation status */
      sos_x86_videomem_printf(2, 0,
			      SOS_X86_VIDEO_FG_YELLOW
			      | SOS_X86_VIDEO_BG_BLUE,
			      "Could allocate %d pages      ",
			      num_alloc_ppages);

      /* We fill this page with its address */
      for (i = 0 ; i < MY_PPAGE_NUM_INT ; i++)
	my_ppage->before[i] = my_ppage->after[i] = (sos_ui32_t)my_ppage;

      /* We add this page at the tail of our list of ppages */
      list_add_tail(ppage_list, my_ppage);
    }

  /* Now we release these pages in FIFO order */
  while ((my_ppage = list_pop_head(ppage_list)) != NULL)
    {
      /* We make sure this page was not overwritten by any unexpected
	 value */
      int i;
      for (i = 0 ; i < MY_PPAGE_NUM_INT ; i++)
	{
	  /* We don't get what we expect ! */
	  if ((my_ppage->before[i] !=  (sos_ui32_t)my_ppage)
	      || (my_ppage->after[i] !=  (sos_ui32_t)my_ppage))
	    {
	      /* STOP ! */
	      sos_x86_videomem_putstring(20, 0,
					 SOS_X86_VIDEO_FG_LTRED
					   | SOS_X86_VIDEO_BG_BLUE,
					 "Page overwritten");
	      return;
	    }
	}

      /* Release the descriptor */
      if (sos_physmem_unref_physpage((sos_paddr_t)my_ppage) < 0)
	{
	  /* STOP ! */
	  sos_x86_videomem_putstring(20, 0,
				     SOS_X86_VIDEO_FG_LTRED
				       | SOS_X86_VIDEO_BG_BLUE,
				     "Cannot release page");
	  return;
	}

      /* Print the deallocation status */
      num_free_ppages ++;
      sos_x86_videomem_printf(2, 0,
			      SOS_X86_VIDEO_FG_YELLOW
			      | SOS_X86_VIDEO_BG_BLUE,
			      "Could free %d pages      ",
			      num_free_ppages);
    }

  /* Print the overall stats */
  sos_x86_videomem_printf(2, 0,
			  SOS_X86_VIDEO_FG_LTGREEN
			  | SOS_X86_VIDEO_BG_BLUE,
			  "Could allocate %d bytes, could free %d bytes     ",
			  num_alloc_ppages << SOS_PAGE_SHIFT,
			  num_free_ppages << SOS_PAGE_SHIFT);

  SOS_ASSERT_FATAL(num_alloc_ppages == num_free_ppages);
}

// [[[]]]
// [[[$article == 4]]]

/* Page fault exception handler */
static void pgflt_ex(int exid)
{
  sos_bochs_printf("Got page fault\n");
  sos_x86_videomem_printf(10, 30,
			  SOS_X86_VIDEO_FG_LTRED | SOS_X86_VIDEO_BG_BLUE,
			  "Got EXPECTED (?) Page fault ! But where ???");
  for (;;) ;
}

static void test_paging(sos_vaddr_t sos_kernel_core_top_vaddr)
{
  /* The (linear) address of the page holding the code we are
     currently executing */
  sos_vaddr_t vpage_code = SOS_PAGE_ALIGN_INF(test_paging);

  /* The new physical page that will hold the code */
  sos_paddr_t ppage_new;

  /* Where this page will be mapped temporarily in order to copy the
     code into it: right after the kernel code/data */
  sos_vaddr_t vpage_tmp = sos_kernel_core_top_vaddr;

  unsigned i;

  /* Bind the page fault exception to one of our routines */
  sos_exception_set_routine(SOS_EXCEPT_PAGE_FAULT,
			    pgflt_ex);

  /*
   * Test 1: move the page where we execute the code elsewhere in
   * physical memory
   */
  sos_x86_videomem_printf(4, 0,
			  SOS_X86_VIDEO_FG_LTGREEN | SOS_X86_VIDEO_BG_BLUE,
			  "Moving current code elsewhere in physical memory:");


  /* Allocate a new physical page */
  ppage_new = sos_physmem_ref_physpage_new(FALSE);
  if (! ppage_new)
    {
      /* STOP ! No memory left */
      sos_x86_videomem_putstring(20, 0,
				 SOS_X86_VIDEO_FG_LTRED
				   | SOS_X86_VIDEO_BG_BLUE,
				 "test_paging : Cannot allocate page");
      return;
    }

  sos_x86_videomem_printf(5, 0,
			  SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			  "Hello from the address 0x%x in physical memory",
			  sos_paging_get_paddr(vpage_code));

  sos_x86_videomem_printf(6, 0,
			  SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			  "Transfer vpage 0x%x: ppage 0x%x -> 0x%x (tmp vpage 0x%x)",
	                  vpage_code,
			  sos_paging_get_paddr(vpage_code),
			  ppage_new,
			  (unsigned)vpage_tmp);

  /* Map the page somewhere (right after the kernel mapping) in order
     to copy the code we are currently executing */
  sos_paging_map(ppage_new, vpage_tmp,
		 FALSE,
		 SOS_VM_MAP_ATOMIC
		 | SOS_VM_MAP_PROT_READ
		 | SOS_VM_MAP_PROT_WRITE);

  /* Ok, the new page is referenced by the mapping, we can release our
     reference to it */
  sos_physmem_unref_physpage(ppage_new);

  /* Copy the contents of the current page of code to this new page
     mapping */
  memcpy((void*)vpage_tmp,
	 (void*)vpage_code,
	 SOS_PAGE_SIZE);

  /* Transfer the mapping of the current page of code to this new page */
  sos_paging_map(ppage_new, vpage_code,
		 FALSE,
		 SOS_VM_MAP_ATOMIC
		 | SOS_VM_MAP_PROT_READ
		 | SOS_VM_MAP_PROT_WRITE);
  
  /* Ok, here we are: we have changed the physcal page that holds the
     code we are executing ;). However, this new page is mapped at 2
     virtual addresses:
     - vpage_tmp
     - vpage_code
     We can safely unmap it from sos_kernel_core_top_vaddr, while
     still keeping the vpage_code mapping */
  sos_paging_unmap(vpage_tmp);

  sos_x86_videomem_printf(7, 0,
			  SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			  "Hello from the address 0x%x in physical memory",
			  sos_paging_get_paddr(vpage_code));

  sos_x86_videomem_printf(9, 0,
			  SOS_X86_VIDEO_FG_LTGREEN | SOS_X86_VIDEO_BG_BLUE,
			  "Provoking a page fault:");

  /*
   * Test 2: make sure the #PF handler works
   */

  /* Scan part of the kernel up to a page fault. This page fault
     should occur on the first page unmapped after the kernel area,
     which is exactly the page we temporarily mapped/unmapped
     (vpage_tmp) above to move the kernel code we are executing */
  for (i = vpage_code ; /* none */ ; i += SOS_PAGE_SIZE)
    {
      unsigned *pint = (unsigned *)SOS_PAGE_ALIGN_INF(i);
      sos_bochs_printf("Test vaddr 0x%x : val=", (unsigned)pint);
      sos_x86_videomem_printf(10, 0,
			      SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			      "Test vaddr 0x%x : val=      ",
			      (unsigned)pint);
      sos_bochs_printf("0x%x\n", *pint);
      sos_x86_videomem_printf(10, 30,
			      SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			      "0x%x          ", *pint);
    }

  /* BAD ! Did not get the page fault... */
  sos_x86_videomem_printf(20, 0,
			  SOS_X86_VIDEO_FG_LTRED | SOS_X86_VIDEO_BG_BLUE,
			  "We should have had a #PF at vaddr 0x%x !",
			  vpage_tmp);
}
// [[[]]]
// [[[$article == 5]]]
struct digit
{
  struct digit *prev, *next;
  char value;
};

/* Representation of a big (positive) integer: Most Significant Digit
   (MSD) is the HEAD of the list. Least Significant Digit (LSD) is the
   TAIL of the list */
typedef struct digit * big_number_t;


/* Add a new digit after the LSD */
void bn_push_lsd(big_number_t * bn, char value)
{
  struct digit *d;
  d = (struct digit*) sos_kmalloc(sizeof(struct digit), 0);
  SOS_ASSERT_FATAL(d != NULL);
  d->value = value;
  list_add_tail(*bn, d);
}


/* Add a new digit before the MSD */
void bn_push_msd(big_number_t * bn, char value)
{
  struct digit *d;
  d = (struct digit*) sos_kmalloc(sizeof(struct digit), 0);
  SOS_ASSERT_FATAL(d != NULL);
  d->value = value;
  list_add_head(*bn, d);
}


/* Construct a big integer from a (machine) integer */
big_number_t bn_new(unsigned long int i)
{
  big_number_t retval;

  list_init(retval);
  do
    {
      bn_push_msd(&retval, i%10);
      i /= 10;
    }
  while (i != 0);

  return retval;
}


/* Create a new big integer from another big integer */
big_number_t bn_copy(const big_number_t bn)
{
  big_number_t retval;
  int nb_elts;
  struct digit *d;

  list_init(retval);
  list_foreach(bn, d, nb_elts)
    {
      bn_push_lsd(&retval, d->value);
    }

  return retval;
}


/* Free the memory used by a big integer */
void bn_del(big_number_t * bn)
{
  struct digit *d;

  list_collapse(*bn, d)
    {
      sos_kfree((sos_vaddr_t)d);
    }
}


/* Shift left a big integer: bn := bn*10^shift */
void bn_shift(big_number_t *bn, int shift)
{
  for ( ; shift > 0 ; shift --)
    {
      bn_push_lsd(bn, 0);
    }
}


/* Dump the big integer in bochs */
void bn_print_bochs(const big_number_t bn)
{
  int nb_elts;
  const struct digit *d;

  if (list_is_empty(bn))
    sos_bochs_printf("0");
  else
    list_foreach(bn, d, nb_elts)
      sos_bochs_printf("%d", d->value);
}

/* Dump the big integer on the console */
void bn_print_console(unsigned char row, unsigned char col,
		      unsigned char attribute,
		      const big_number_t bn,
		      int nb_decimals)
{
  if (list_is_empty(bn))
    sos_x86_videomem_printf(row, col, attribute, "0");
  else
    {
      int nb_elts;
      const struct digit *d;
      unsigned char x = col;

      list_foreach(bn, d, nb_elts)
	{
	  if (nb_elts == 0)
	    {
	      sos_x86_videomem_printf(row, x, attribute, "%d.", d->value);
	      x += 2;
	    }
	  else if (nb_elts < nb_decimals)
	    {
	      sos_x86_videomem_printf(row, x, attribute, "%d", d->value);
	      x ++;
	    }
	}

      sos_x86_videomem_printf(row, x, attribute, " . 10^{%d}  ", nb_elts-1);
    }
}


/* Result is the addition of 2 big integers */
big_number_t bn_add (const big_number_t bn1, const big_number_t bn2)
{
  big_number_t retval;
  const struct digit *d1, *d2;
  sos_bool_t  bn1_end = FALSE, bn2_end = FALSE;
  char carry = 0;

  list_init(retval);
  d1 = list_get_tail(bn1);
  bn1_end = list_is_empty(bn1);
  d2 = list_get_tail(bn2);
  bn2_end = list_is_empty(bn2);
  do
    {
      if (! bn1_end)
	carry += d1->value;
      if (! bn2_end)
	carry += d2->value;

      bn_push_msd(&retval, carry % 10);
      carry  /= 10;

      if (! bn1_end)
	d1 = d1->prev;
      if (! bn2_end)
	d2 = d2->prev;
      if (d1 == list_get_tail(bn1))
	bn1_end = TRUE;
      if (d2 == list_get_tail(bn2))
	bn2_end = TRUE;
    }
  while (!bn1_end || !bn2_end);

  if (carry > 0)
    {
      bn_push_msd(&retval, carry);
    }

  return retval;
}


/* Result is the multiplication of a big integer by a single digit */
big_number_t bn_muli (const big_number_t bn, char digit)
{
  big_number_t retval;
  int nb_elts;
  char   carry = 0;
  const struct digit *d;

  list_init(retval);
  list_foreach_backward(bn, d, nb_elts)
    {
      carry += d->value * digit;
      bn_push_msd(&retval, carry % 10);
      carry /= 10;
    }

  if (carry > 0)
    {
      bn_push_msd(&retval, carry);
    }

  return retval;
}


/* Result is the multiplication of 2 big integers */
big_number_t bn_mult(const big_number_t bn1, const big_number_t bn2)
{
  int shift = 0;
  big_number_t retval;
  int nb_elts;
  struct digit *d;

  list_init(retval);
  list_foreach_backward(bn2, d, nb_elts)
    {
      big_number_t retmult = bn_muli(bn1, d->value);
      big_number_t old_retval = retval;
      bn_shift(& retmult, shift);
      retval = bn_add(old_retval, retmult);
      bn_del(& retmult);
      bn_del(& old_retval);
      shift ++;
    }

  return retval;
}


/* Result is the factorial of an integer */
big_number_t bn_fact(unsigned long int v)
{
  unsigned long int i;
  big_number_t retval = bn_new(1);
  for (i = 1 ; i <= v ; i++)
    {
      big_number_t I   = bn_new(i);
      big_number_t tmp = bn_mult(retval, I);
      sos_x86_videomem_printf(4, 0,
			      SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_LTGREEN,
			      "%d! = ", (int)i);
      bn_print_console(4, 8, SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_WHITE,
		       tmp, 55);
      bn_del(& I);
      bn_del(& retval);
      retval = tmp;
    }

  return retval;
}


void bn_test()
{
  big_number_t bn = bn_fact(1000);
  sos_bochs_printf("1000! = ");
  bn_print_bochs(bn);
  sos_bochs_printf("\n");
  
}

// [[[]]]
// [[[$article == 6]]]


/* ======================================================================
 * Demonstrate the use of the CPU kernet context management API:
 *  - A coroutine prints "Hlowrd" and switches to the other after each
 *    letter
 *  - A coroutine prints "el ol\n" and switches back to the other after
 *    each letter.
 * The first to reach the '\n' returns back to main.
 */
struct sos_cpu_state *ctxt_hello1;
struct sos_cpu_state *ctxt_hello2;
struct sos_cpu_state *ctxt_main;
sos_vaddr_t hello1_stack, hello2_stack;

static void reclaim_stack(sos_vaddr_t stack_vaddr)
{
  sos_kfree(stack_vaddr);
}


static void exit_hello12(sos_vaddr_t stack_vaddr)
{
  sos_cpu_context_exit_to(ctxt_main,
			  (sos_cpu_kstate_function_arg1_t*) reclaim_stack,
			  stack_vaddr);
}


static void hello1 (char *str)
{
  for ( ; *str != '\n' ; str++)
    {
      sos_bochs_printf("hello1: %c\n", *str);
      sos_cpu_context_switch(& ctxt_hello1, ctxt_hello2);
    }

  /* You can uncomment this in case you explicitly want to exit
     now. But returning from the function will do the same */
  /* sos_cpu_context_exit_to(ctxt_main,
			     (sos_cpu_kstate_function_arg1_t*) reclaim_stack,
			     hello1_stack); */
}


static void hello2 (char *str)
{
  for ( ; *str != '\n' ; str++)
    {
      sos_bochs_printf("hello2: %c\n", *str);
      sos_cpu_context_switch(& ctxt_hello2, ctxt_hello1);
    }

  /* You can uncomment this in case you explicitly want to exit
     now. But returning from the function will do the same */
  /* sos_cpu_context_exit_to(ctxt_main,
			     (sos_cpu_kstate_function_arg1_t*) reclaim_stack,
			     hello2_stack); */
}


void print_hello_world ()
{
#define DEMO_STACK_SIZE 1024
  /* Allocate the stacks */
  hello1_stack = sos_kmalloc(DEMO_STACK_SIZE, 0);
  hello2_stack = sos_kmalloc(DEMO_STACK_SIZE, 0);

  /* Initialize the coroutines' contexts */
  sos_cpu_kstate_init(&ctxt_hello1,
                      (sos_cpu_kstate_function_arg1_t*) hello1,
		      (sos_ui32_t) "Hlowrd",
                      (sos_vaddr_t) hello1_stack, DEMO_STACK_SIZE,
                      (sos_cpu_kstate_function_arg1_t*) exit_hello12,
		      (sos_ui32_t) hello1_stack);
  sos_cpu_kstate_init(&ctxt_hello2,
                      (sos_cpu_kstate_function_arg1_t*) hello2,
		      (sos_ui32_t) "el ol\n",
                      (sos_vaddr_t) hello2_stack, DEMO_STACK_SIZE,
                      (sos_cpu_kstate_function_arg1_t*) exit_hello12,
		      (sos_ui32_t) hello2_stack);

  /* Go to first coroutine */
  sos_bochs_printf("Printing Hello World\\n...\n");
  sos_cpu_context_switch(& ctxt_main, ctxt_hello1);

  /* The first coroutine to reach the '\n' switched back to us */
  sos_bochs_printf("Back in main !\n");
}


/* ======================================================================
 * Generate page faults on an unmapped but allocated kernel virtual
 * region, which results in a series of physical memory mappings for the
 * faulted pages.
 */
static void test_demand_paging(int nb_alloc_vpages, int nb_alloc_ppages)
{
  int i;
  sos_vaddr_t base_vaddr;

  sos_x86_videomem_printf(10, 0,
			  SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_LTGREEN,
			  "Demand paging test (alloc %dMB of VMM, test %dkB RAM)",
			  nb_alloc_vpages >> 8, nb_alloc_ppages << 2);
  
  /* Allocate virtual memory */
  base_vaddr = sos_kmem_vmm_alloc(nb_alloc_vpages, 0);

  SOS_ASSERT_FATAL(base_vaddr != (sos_vaddr_t)NULL);
  sos_x86_videomem_printf(11, 0,
			  SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_YELLOW,
			  "Allocated virtual region [0x%x, 0x%x[",
			  base_vaddr,
			  base_vaddr + nb_alloc_vpages*SOS_PAGE_SIZE);

  /* Now use part of it in physical memory */
  for (i = 0 ; (i < nb_alloc_ppages) && (i < nb_alloc_vpages) ; i++)
    {
      /* Compute an address inside the range */
      sos_ui32_t *value, j;
      sos_vaddr_t vaddr = base_vaddr;
      vaddr += (nb_alloc_vpages - (i + 1))*SOS_PAGE_SIZE;
      vaddr += 2345;

      sos_x86_videomem_printf(12, 0,
			      SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_YELLOW,
			      "Writing %d at virtual address 0x%x...",
			      i, vaddr);

      /* Write at this address */
      value = (sos_ui32_t*)vaddr;
      *value = i;

      /* Yep ! A new page should normally have been allocated for us */
      sos_x86_videomem_printf(13, 0,
			      SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_YELLOW,
			      "Value read at address 0x%x = %d",
			      vaddr, (unsigned)*value);
    }

  SOS_ASSERT_FATAL(SOS_OK == sos_kmem_vmm_free(base_vaddr));
  /* Yep ! A new page should normally have been allocated for us */
  sos_x86_videomem_printf(14, 0,
			  SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_YELLOW,
			  "Done (area un-allocated)");
}



/* ======================================================================
 * Shows how the backtrace stuff works
 */

/* Recursive function. Print the backtrace from the innermost function */
static void test_backtrace(int i, int magic, sos_vaddr_t stack_bottom,
			   sos_size_t  stack_size)
{
  if (i <= 0)
    {
      /* The page fault exception handler will print the backtrace of
	 this function, because address 0x42 is not mapped */
      *((char*)0x42) = 12;

      /* More direct variant: */
      /* dump_backtrace(NULL, stack_bottom, stack_size, TRUE, TRUE); */
    }
  else
    test_backtrace(i-1, magic, stack_bottom, stack_size);
}


/* ======================================================================
 * Parsing of Mathematical expressions
 *
 * This is a recursive lexer/parser/evaluator for arithmetical
 * expressions. Supports both binary +/-* and unary +- operators, as
 * well as parentheses.
 *
 * Terminal tokens (Lexer):
 *  - Number: positive integer number
 *  - Variable: ascii name (regexp: [a-zA-Z]+)
 *  - Operator: +*-/
 *  - Opening/closing parentheses
 *
 * Grammar (Parser):
 *  Expression ::= Term E'
 *  Expr_lr    ::= + Term Expr_lr | - Term Expr_lr | Nothing
 *  Term       ::= Factor Term_lr
 *  Term_lr    ::= * Factor Term_lr | / Factor Term_lr | Nothing
 *  Factor     ::= - Factor | + Factor | Scalar | ( Expression )
 *  Scalar     ::= Number | Variable
 *
 * Note. This is the left-recursive equivalent of the following basic grammar:
 *  Expression ::= Expression + Term | Expression - Term
 *  Term       ::= Term * Factor | Term / Factor
 *  factor     ::= - Factor | + Factor | Scalar | Variable | ( Expression )
 *  Scalar     ::= Number | Variable
 *
 * The parsing is composed of a 3 stages pipeline:
 *  - The reader: reads a string 1 character at a time, transferring
 *    the control back to lexer after each char. This function shows the
 *    interest in using coroutines, because its state (str) is
 *    implicitely stored in the stack between each iteration.
 *  - The lexer: consumes the characters from the reader and identifies
 *    the terminal tokens, 1 token at a time, transferring control back
 *    to the parser after each token. This function shows the interest
 *    in using coroutines, because its state (c and got_what_before) is
 *    implicitely stored in the stack between each iteration.
 *  - The parser: consumes the tokens from the lexer and builds the
 *    syntax tree of the expression. There is no real algorithmic
 *    interest in defining a coroutine devoted to do this. HOWEVER, we
 *    do use one for that because this allows us to switch to a much
 *    deeper stack. Actually, the parser is highly recursive, so that
 *    the default 16kB stack of the sos_main() function might not be
 *    enough. Here, we switch to a 64kB stack, which is safer for
 *    recursive functions. The Parser uses intermediary functions: these
 *    are defined and implemented as internal nested functions. This is
 *    just for the sake of clarity, and is absolutely not mandatory for
 *    the algorithm: one can transfer these functions out of the parser
 *    function without restriction.
 *
 * The evaluator is another recursive function that reuses the
 * parser's stack to evaluate the parsed expression with the given
 * values for the variables present in the expression. As for the
 * parser function, this function defines and uses a nested function,
 * which can be extracted from the main evaluation function at will.
 *
 * All these functions support a kind of "exception" feature: when
 * something goes wrong, control is transferred DIRECTLY back to the
 * sos_main() context, without unrolling the recursions. This shows
 * how exceptions basically work, but one should not consider this as
 * a reference exceptions implementation. Real exception mechanisms
 * (such as that in the C++ language) call the destructors to the
 * objects allocated on the stack during the "stack unwinding" process
 * upon exception handling, which complicates a lot the mechanism. We
 * don't have real Objects here (in the OOP sense, full-featured with
 * destructors), so we don't have to complicate things.
 *
 * After this little coroutine demo, one should forget all about such
 * a low-level manual direct manipulation of stacks. This would
 * probably mess up the whole kernel to do what we do here (locked
 * resources such as mutex/semaphore won't be correctly unlocked,
 * ...). Higher level "kernel thread" primitives will soon be
 * presented, which provide a higher-level set of APIs to manage CPU
 * contexts. You'll have to use EXCLUSIVELY those APIs. If you still
 * need a huge stack to do recursion for example, please don't even
 * think of changing manually the stack for something bigger ! Simply
 * rethink your algorithm, making it non-recursive.
 */


/* The stacks involved */
static char stack_reader[1024];
static char stack_lexer[1024];
static char deep_stack[65536]; /* For the parser and the evaluator */

/* The CPU states for the various coroutines */
static struct sos_cpu_state *st_reader, *st_lexer, *st_parser,
  *st_eval, *st_free, *st_main;


/*
 * Default exit/reclaim functions: return control to the "sos_main()"
 * context
 */
static void reclaim(int unused)
{
}
static void func_exit(sos_ui32_t unused)
{
  sos_cpu_context_exit_to(st_main, (sos_cpu_kstate_function_arg1_t*)reclaim, 0);
}


/*
 * The reader coroutine and associated variable. This coroutine could
 * have been a normal function, except that the current parsed
 * character would have to be stored somewhere.
 */
static char data_reader_to_lexer;

static void func_reader(const char *str)
{
  for ( ; str && (*str != '\0') ; str++)
    {
      data_reader_to_lexer = *str;
      sos_cpu_context_switch(& st_reader, st_lexer);
    }

  data_reader_to_lexer = '\0';
  sos_cpu_context_switch(& st_reader, st_lexer);
}


/*
 * The Lexer coroutine and associated types/variables. This coroutine
 * could have been a normal function, except that the current parsed
 * character, token and previous token would have to be stored
 * somewhere.
 */
#define STR_VAR_MAXLEN 16
static struct lex_elem
{
  enum { LEX_IS_NUMBER, LEX_IS_OPER, LEX_IS_VAR,
	 LEX_IS_OPENPAR, LEX_IS_CLOSEPAR, LEX_END } type;
  union {
    int  number;
    char operator;
    char var[STR_VAR_MAXLEN];
  };
} data_lexer_to_parser;

static void func_lexer(sos_ui32_t unused)
{
  char c;
  enum { GOT_SPACE, GOT_NUM, GOT_OP, GOT_STR,
	 GOT_OPENPAR, GOT_CLOSEPAR } got_what, got_what_before;

  data_lexer_to_parser.number = 0;
  got_what_before = GOT_SPACE;
  do
    {
      /* Consume one character from the reader */
      sos_cpu_context_switch(& st_lexer, st_reader);
      c = data_reader_to_lexer;

      /* Classify the consumed character */
      if ( (c >= '0') && (c <= '9') )
	got_what = GOT_NUM;
      else if ( (c == '+') || (c == '-') || (c == '*') || (c == '/') )
	got_what = GOT_OP;
      else if ( ( (c >= 'a') && (c <= 'z') )
		|| ( (c >= 'A') && (c <= 'Z') ) )
	got_what = GOT_STR;
      else if (c == '(')
	got_what = GOT_OPENPAR;
      else if (c == ')')
	got_what = GOT_CLOSEPAR;
      else
	got_what = GOT_SPACE;

      /* Determine whether the current token is ended */
      if ( (got_what != got_what_before)
	   || (got_what_before == GOT_OP)
	   || (got_what_before == GOT_OPENPAR)
	   || (got_what_before == GOT_CLOSEPAR) )
	{
	  // [[[$master && $article==6 ]]]
#if 0
	  switch (got_what_before)
	    {
	    case GOT_CLOSEPAR:
	      sos_bochs_printf("LEX: close par\n"); break;
	    case GOT_OPENPAR:
	      sos_bochs_printf("LEX: open par\n"); break;
	    case GOT_STR:
	      sos_bochs_printf("LEX: var %s\n", data_lexer_to_parser.var);
	      break;
	    case GOT_NUM:
	      sos_bochs_printf("LEX: num %d\n", data_lexer_to_parser.number);
	      break;
	    case GOT_OP:
	      sos_bochs_printf("LEX: op %c\n", data_lexer_to_parser.operator);
	      break;
	    case GOT_SPACE:
	      /* Ignore space(s) */
	      break;
	    default:
	      SOS_ASSERT_FATAL(! "Ivalid character ?!");
	    }
#endif
	  // [[[]]]
	  /* return control back to the parser if the previous token
	     has been recognized */
	  if ( (got_what_before != GOT_SPACE) )
	    sos_cpu_context_switch(& st_lexer, st_parser);

	  data_lexer_to_parser.number = 0;
	}

      /* Update the token being currently recognized */
      if (got_what == GOT_OP)
	{
	  data_lexer_to_parser.type = LEX_IS_OPER;
	  data_lexer_to_parser.operator = c;
	}
      else if (got_what == GOT_NUM)
	{
	  data_lexer_to_parser.type = LEX_IS_NUMBER;
	  data_lexer_to_parser.number *= 10;
	  data_lexer_to_parser.number += (c - '0');
	}
      else if (got_what == GOT_STR)
	{
	  char to_cat[] = { c, '\0' };
	  data_lexer_to_parser.type = LEX_IS_VAR;
	  strzcat(data_lexer_to_parser.var, to_cat, STR_VAR_MAXLEN);
	}
      else if (got_what == GOT_OPENPAR)
	data_lexer_to_parser.type = LEX_IS_OPENPAR;
      else if (got_what == GOT_CLOSEPAR)
	data_lexer_to_parser.type = LEX_IS_CLOSEPAR;

      got_what_before = got_what;
    }
  while (c != '\0');

  /* Transfer last recognized token to the parser */
  if ( (got_what_before != GOT_SPACE) )
    sos_cpu_context_switch(& st_lexer, st_parser);

  /* Signal that no more token are available */
  data_lexer_to_parser.type = LEX_END;
  sos_cpu_context_switch(& st_lexer, st_parser);

  /* Exception: parser asks for a token AFTER having received the last
     one */
  sos_bochs_printf("Error: end of string already reached !\n");
  sos_cpu_context_switch(& st_lexer, st_main);
}


/*
 * The Parser coroutine and associated types/variables
 */
struct syntax_node
{
  enum { YY_IS_BINOP, YY_IS_UNAROP, YY_IS_NUM, YY_IS_VAR } type;
  union
  {
    int  number;
    char var[STR_VAR_MAXLEN];
    struct
    {
      char op;
      struct syntax_node *parm_left, *parm_right;
    } binop;
    struct
    {
      char op;
      struct syntax_node *parm;
    } unarop;
  };
};

static void func_parser(struct syntax_node ** syntax_tree)
{
  static struct syntax_node *alloc_node_num(int val);
  static struct syntax_node *alloc_node_var(const char * name);
  static struct syntax_node *alloc_node_binop(char op,
					      struct syntax_node *parm_left,
					      struct syntax_node *parm_right);
  static struct syntax_node *alloc_node_unarop(char op,
					       struct syntax_node *parm);
  static struct syntax_node * get_expr();
  static struct syntax_node * get_expr_lr(struct syntax_node *n);
  static struct syntax_node * get_term();
  static struct syntax_node * get_term_lr(struct syntax_node *n);
  static struct syntax_node * get_factor();
  static struct syntax_node * get_scalar();

  /* Create a new node to store a number */
  static struct syntax_node *alloc_node_num(int val)
    {
      struct syntax_node *n
	= (struct syntax_node*) sos_kmalloc(sizeof(struct syntax_node), 0);
      n->type   = YY_IS_NUM;
      n->number = val;
      return n;
    }
  /* Create a new node to store a variable */
  static struct syntax_node *alloc_node_var(const char * name)
    {
      struct syntax_node *n
	= (struct syntax_node*) sos_kmalloc(sizeof(struct syntax_node), 0);
      n->type   = YY_IS_VAR;
      strzcpy(n->var, name, STR_VAR_MAXLEN);
      return n;
    }
  /* Create a new node to store a binary operator */
  static struct syntax_node *alloc_node_binop(char op,
					      struct syntax_node *parm_left,
					      struct syntax_node *parm_right)
    {
      struct syntax_node *n
	= (struct syntax_node*) sos_kmalloc(sizeof(struct syntax_node), 0);
      n->type             = YY_IS_BINOP;
      n->binop.op         = op;
      n->binop.parm_left  = parm_left;
      n->binop.parm_right = parm_right;
      return n;
    }
  /* Create a new node to store a unary operator */
  static struct syntax_node *alloc_node_unarop(char op,
					       struct syntax_node *parm)
    {
      struct syntax_node *n
	= (struct syntax_node*) sos_kmalloc(sizeof(struct syntax_node), 0);
      n->type        = YY_IS_UNAROP;
      n->unarop.op   = op;
      n->unarop.parm = parm;
      return n;
    }

  /* Raise an exception: transfer control back to main context,
     without unrolling the whole recursion */
  static void parser_exception(const char *str)
    {
      sos_bochs_printf("Parser exception: %s\n", str);
      sos_cpu_context_switch(& st_parser, st_main);
    }

  /* Consume the current terminal "number" token and ask for a new
     token */
  static int get_number()
    {
      int v;
      if (data_lexer_to_parser.type != LEX_IS_NUMBER)
	parser_exception("Expected number");
      v = data_lexer_to_parser.number;
      sos_cpu_context_switch(& st_parser, st_lexer);
      return v;
    }
  /* Consume the current terminal "variable" token and ask for a new
     token */
  static void get_str(char name[STR_VAR_MAXLEN])
    {
      if (data_lexer_to_parser.type != LEX_IS_VAR)
	parser_exception("Expected variable");
      strzcpy(name, data_lexer_to_parser.var, STR_VAR_MAXLEN);
      sos_cpu_context_switch(& st_parser, st_lexer);
    }
  /* Consume the current terminal "operator" token and ask for a new
     token */
  static char get_op()
    {
      char op;
      if (data_lexer_to_parser.type != LEX_IS_OPER)
	parser_exception("Expected operator");
      op = data_lexer_to_parser.operator;
      sos_cpu_context_switch(& st_parser, st_lexer);
      return op;
    }
  /* Consume the current terminal "parenthese" token and ask for a new
     token */
  static void get_par()
    {
      if ( (data_lexer_to_parser.type != LEX_IS_OPENPAR)
	   && (data_lexer_to_parser.type != LEX_IS_CLOSEPAR) )
	parser_exception("Expected parenthese");
      sos_cpu_context_switch(& st_parser, st_lexer);
    }

  /* Parse an Expression */
  static struct syntax_node * get_expr()
    {
      struct syntax_node *t = get_term();
      return get_expr_lr(t);
    }
  /* Parse an Expr_lr */
  static struct syntax_node * get_expr_lr(struct syntax_node *n)
    {
      if ( (data_lexer_to_parser.type == LEX_IS_OPER)
	   && ( (data_lexer_to_parser.operator == '+')
		|| (data_lexer_to_parser.operator == '-') ) )
	{
	  char op = get_op();
	  struct syntax_node *term = get_term();
	  struct syntax_node *node_op = alloc_node_binop(op, n, term);
	  return get_expr_lr(node_op);
	}
      return n;
    }
  /* Parse a Term */
  static struct syntax_node * get_term()
    {
      struct syntax_node *f1 = get_factor();
      return get_term_lr(f1);
    }
  /* Parse a Term_lr */
  static struct syntax_node * get_term_lr(struct syntax_node *n)
    {
      if ( (data_lexer_to_parser.type == LEX_IS_OPER)
	   && ( (data_lexer_to_parser.operator == '*')
		|| (data_lexer_to_parser.operator == '/') ) )
	{
	  char op = get_op();
	  struct syntax_node *factor = get_factor();
	  struct syntax_node *node_op = alloc_node_binop(op, n, factor);
	  return get_term_lr(node_op);
	}
      return n;
    }
  /* Parse a Factor */
  static struct syntax_node * get_factor()
    {
      if ( (data_lexer_to_parser.type == LEX_IS_OPER)
	   && ( (data_lexer_to_parser.operator == '-')
		|| (data_lexer_to_parser.operator == '+') ) )
	{ char op = data_lexer_to_parser.operator;
	get_op(); return alloc_node_unarop(op, get_factor()); }
      else if (data_lexer_to_parser.type == LEX_IS_OPENPAR)
	{
	  struct syntax_node *expr;
	  get_par();
	  expr = get_expr();
	  if (data_lexer_to_parser.type != LEX_IS_CLOSEPAR)
	    parser_exception("Mismatched parentheses");
	  get_par();
	  return expr;
	}
  
      return get_scalar();
    }
  /* Parse a Scalar */
  static struct syntax_node * get_scalar()
    {
      if (data_lexer_to_parser.type != LEX_IS_NUMBER)
	{
	  char var[STR_VAR_MAXLEN];
	  get_str(var);
	  return alloc_node_var(var);
	}
      return alloc_node_num(get_number());
    }


  /*
   * Body of the function
   */

  /* Get the first token */
  sos_cpu_context_switch(& st_parser, st_lexer);

  /* Begin the parsing ! */
  *syntax_tree = get_expr();
  /* The result is returned in the syntax_tree parameter */
}


/*
 * Setup the parser's pipeline
 */
static struct syntax_node * parse_expression(const char *expr)
{
  struct syntax_node *retval = NULL;

  /* Build the context of the functions in the pipeline */
  sos_cpu_kstate_init(& st_reader,
		      (sos_cpu_kstate_function_arg1_t*)func_reader,
		      (sos_ui32_t)expr,
		      (sos_vaddr_t)stack_reader, sizeof(stack_reader),
		      (sos_cpu_kstate_function_arg1_t*)func_exit, 0);
  sos_cpu_kstate_init(& st_lexer,
		      (sos_cpu_kstate_function_arg1_t*)func_lexer,
		      0,
		      (sos_vaddr_t)stack_lexer, sizeof(stack_lexer),
		      (sos_cpu_kstate_function_arg1_t*)func_exit, 0);
  sos_cpu_kstate_init(& st_parser,
		      (sos_cpu_kstate_function_arg1_t*)func_parser,
		      (sos_ui32_t) /* syntax tree ! */&retval,
		      (sos_vaddr_t)deep_stack, sizeof(deep_stack),
		      (sos_cpu_kstate_function_arg1_t*)func_exit, 0);

  /* Parse the expression */
  sos_cpu_context_switch(& st_main, st_parser);
  return retval;
}


/*
 * The Evaluator coroutine and associated types/variables
 */
struct func_eval_params
{
  const struct syntax_node *e;
  const char **var_name;
  int *var_val;
  int nb_vars;

  int result;
};

static void func_eval(struct func_eval_params *parms)
{
  /* The internal (recursive) nested function to evaluate each node of
     the syntax tree */
  static int rec_eval(const struct syntax_node *n,
		      const char* var_name[], int var_val[], int nb_vars)
    {
      switch (n->type)
	{
	case YY_IS_NUM:
	  return n->number;

	case YY_IS_VAR:
	  {
	    int i;
	    for (i = 0 ; i < nb_vars ; i++)
	      if (0 == strcmp(var_name[i], n->var))
		return var_val[i];

	    /* Exception: no variable with that name ! */
	    sos_bochs_printf("ERROR: unknown variable %s\n", n->var);
	    sos_cpu_context_switch(& st_eval, st_main);
	  }

	case YY_IS_BINOP:
	  {
	    int left = rec_eval(n->binop.parm_left,
				var_name, var_val, nb_vars);
	    int right = rec_eval(n->binop.parm_right,
				 var_name, var_val, nb_vars);
	    switch (n->binop.op)
	      {
	      case '+': return left + right;
	      case '-': return left - right;
	      case '*': return left * right;
	      case '/': return left / right;
	      default:
		/* Exception: no such operator (INTERNAL error) ! */
		sos_bochs_printf("ERROR: unknown binop %c\n", n->binop.op);
		sos_cpu_context_switch(& st_eval, st_main);
	      }
	  }

	case YY_IS_UNAROP:
	  {
	    int arg = rec_eval(n->unarop.parm, var_name, var_val, nb_vars);
	    switch (n->unarop.op)
	      {
	      case '-': return -arg;
	      case '+': return arg;
	      default:
		/* Exception: no such operator (INTERNAL error) ! */
		sos_bochs_printf("ERROR: unknown unarop %c\n", n->unarop.op);
		sos_cpu_context_switch(& st_eval, st_main);
	      }
	  }
	}
      
      /* Exception: no such syntax node (INTERNAL error) ! */
      sos_bochs_printf("ERROR: invalid node type\n");
      sos_cpu_context_switch(& st_eval, st_main);
      return -1; /* let's make gcc happy */
    }


  /*
   * Function BODY
   */
  /* Update p.result returned back to calling function */
  parms->result
    = rec_eval(parms->e, parms->var_name, parms->var_val, parms->nb_vars);
}

/*
 * Change the stack for something larger in order to call the
 * recursive function above in a safe way
 */
static int eval_expression(const struct syntax_node *e,
			   const char* var_name[], int var_val[], int nb_vars)
{
  struct func_eval_params p
    = (struct func_eval_params){ .e=e,
				 .var_name=var_name,
				 .var_val=var_val,
				 .nb_vars=nb_vars,
				 .result = 0 };

  sos_cpu_kstate_init(& st_eval,
		      (sos_cpu_kstate_function_arg1_t*)func_eval,
		      (sos_ui32_t)/* p.result is modified upon success */&p,
		      (sos_vaddr_t)deep_stack, sizeof(deep_stack),
		      (sos_cpu_kstate_function_arg1_t*)func_exit, 0);

  /* Go ! */
  sos_cpu_context_switch(& st_main, st_eval);
  return p.result;
}


/*
 * Function to free the syntax tree
 */
static void func_free(struct syntax_node *n)
{
  switch (n->type)
    {
    case YY_IS_NUM:
    case YY_IS_VAR:
      break;
      
    case YY_IS_BINOP:
      func_free(n->binop.parm_left);
      func_free(n->binop.parm_right);
      break;
      
    case YY_IS_UNAROP:
      func_free(n->unarop.parm);
      break;
    }
  
  sos_kfree((sos_vaddr_t)n);
}

/*
 * Change the stack for something larger in order to call the
 * recursive function above in a safe way
 */
static void free_syntax_tree(struct syntax_node *tree)
{
  sos_cpu_kstate_init(& st_free,
		      (sos_cpu_kstate_function_arg1_t*)func_free,
		      (sos_ui32_t)tree,
		      (sos_vaddr_t)deep_stack, sizeof(deep_stack),
		      (sos_cpu_kstate_function_arg1_t*)func_exit, 0);

  /* Go ! */
  sos_cpu_context_switch(& st_main, st_free);
}
// [[[]]]
// [[[$article >= 6.5 && ($article < 7.5)]]]

/* ======================================================================
 * Demonstrate the use of SOS kernel threads
 *  - Kernel Threads are created with various priorities and their
 *    state is printed on both the console and the bochs' 0xe9 port
 *  - For tests regarding threads' synchronization, see mouse_sim.c
 */

struct thr_arg
{
  char character;
  int  color;

  int col;
  int row;
};


static void demo_thread(void *arg)
{
  struct thr_arg *thr_arg = (struct thr_arg*)arg;
  int progress = 0;

  sos_bochs_printf("start %c", thr_arg->character);
  while (1)
    {
      progress ++;
      display_bits(thr_arg->row, thr_arg->col+1, thr_arg->color, progress);

      sos_bochs_putchar(thr_arg->character);

      /* Yield the CPU to another thread sometimes... */
      if ((random() % 100) == 0)
	{
	  sos_bochs_printf("[37myield(%c)[m\n", thr_arg->character);
	  sos_x86_videomem_putchar(thr_arg->row, thr_arg->col, 0x1e, 'Y');
	  SOS_ASSERT_FATAL(SOS_OK == sos_thread_yield());
	  sos_x86_videomem_putchar(thr_arg->row, thr_arg->col, 0x1e, 'R');
	}

      /* Go to sleep some other times... */
      else if ((random() % 200) == 0)
	{
	  struct sos_time t = (struct sos_time){ .sec=0, .nanosec=50000000 };
	  sos_bochs_printf("[37msleep1(%c)[m\n", thr_arg->character);
	  sos_x86_videomem_putchar(thr_arg->row, thr_arg->col, 0x1e, 's');
	  SOS_ASSERT_FATAL(SOS_OK == sos_thread_sleep(& t));
	  SOS_ASSERT_FATAL(sos_time_is_zero(& t));
	  sos_x86_videomem_putchar(thr_arg->row, thr_arg->col, 0x1e, 'R');
	}

      /* Go to sleep for a longer time some other times... */
      else if ((random() % 300) == 0)
	{
	  struct sos_time t = (struct sos_time){ .sec=0, .nanosec=300000000 };
	  sos_bochs_printf("[37msleep2(%c)[m\n", thr_arg->character);
	  sos_x86_videomem_putchar(thr_arg->row, thr_arg->col, 0x1e, 'S');
	  SOS_ASSERT_FATAL(SOS_OK == sos_thread_sleep(& t));
	  SOS_ASSERT_FATAL(sos_time_is_zero(& t));
	  sos_x86_videomem_putchar(thr_arg->row, thr_arg->col, 0x1e, 'R');
	}

      /* Infinite loop otherwise */
    }
}


static void test_thread()
{
  /* "static" variables because we want them to remain even when the
     function returns */
  static struct thr_arg arg_b, arg_c, arg_d, arg_e, arg_R, arg_S;
  sos_ui32_t flags;

  sos_disable_IRQs(flags);

// [[[$enclosing && $article < 6.75]]]
  arg_b = (struct thr_arg) { .character='b', .col=0, .row=21, .color=0x14 };
  sos_create_kernel_thread("YO[b]", demo_thread, (void*)&arg_b);

  arg_c = (struct thr_arg) { .character='c', .col=46, .row=21, .color=0x14 };
  sos_create_kernel_thread("YO[c]", demo_thread, (void*)&arg_c);

  arg_d = (struct thr_arg) { .character='d', .col=0, .row=20, .color=0x14 };
  sos_create_kernel_thread("YO[d]", demo_thread, (void*)&arg_d);

  arg_e = (struct thr_arg) { .character='e', .col=0, .row=19, .color=0x14 };
  sos_create_kernel_thread("YO[e]", demo_thread, (void*)&arg_e);

  arg_R = (struct thr_arg) { .character='R', .col=0, .row=17, .color=0x1c };
  sos_create_kernel_thread("YO[R]", demo_thread, (void*)&arg_R);

  arg_S = (struct thr_arg) { .character='S', .col=0, .row=16, .color=0x1c };
  sos_create_kernel_thread("YO[S]", demo_thread, (void*)&arg_S);
// [[[]]]
// [[[$enclosing && $article >= 6.75]]]
  arg_b = (struct thr_arg) { .character='b', .col=0, .row=21, .color=0x14 };
  sos_create_kernel_thread("YO[b]", demo_thread, (void*)&arg_b, SOS_SCHED_PRIO_TS_LOWEST);

  arg_c = (struct thr_arg) { .character='c', .col=46, .row=21, .color=0x14 };
  sos_create_kernel_thread("YO[c]", demo_thread, (void*)&arg_c, SOS_SCHED_PRIO_TS_LOWEST);

  arg_d = (struct thr_arg) { .character='d', .col=0, .row=20, .color=0x14 };
  sos_create_kernel_thread("YO[d]", demo_thread, (void*)&arg_d, SOS_SCHED_PRIO_TS_LOWEST-1);

  arg_e = (struct thr_arg) { .character='e', .col=0, .row=19, .color=0x14 };
  sos_create_kernel_thread("YO[e]", demo_thread, (void*)&arg_e, SOS_SCHED_PRIO_TS_LOWEST-2);

  arg_R = (struct thr_arg) { .character='R', .col=0, .row=17, .color=0x1c };
  sos_create_kernel_thread("YO[R]", demo_thread, (void*)&arg_R, SOS_SCHED_PRIO_RT_LOWEST);

  arg_S = (struct thr_arg) { .character='S', .col=0, .row=16, .color=0x1c };
  sos_create_kernel_thread("YO[S]", demo_thread, (void*)&arg_S, SOS_SCHED_PRIO_RT_LOWEST-1);
// [[[]]]

  sos_restore_IRQs(flags);
}
// [[[]]]

// [[[$article >= 6.5]]]

/* ======================================================================
 * An operating system MUST always have a ready thread ! Otherwise:
 * what would the CPU have to execute ?!
 */
static void idle_thread(void* unused) __attribute__((noreturn));
static void idle_thread(void* unused)
{
  sos_ui32_t idle_twiddle = 0;

  while (1)
    {
      /* Remove this instruction if you get an "Invalid opcode" CPU
	 exception (old 80386 CPU) */
      asm("hlt\n");

      idle_twiddle ++;
      display_bits(0, 0, SOS_X86_VIDEO_FG_GREEN | SOS_X86_VIDEO_BG_BLUE,
		   idle_twiddle);
      
      /* Lend the CPU to some other thread */
      sos_thread_yield();
    }
}

// [[[]]]
// [[[$article >= 7]]]

/* ======================================================================
 * Kernel thread showing some CPU usage statistics on the console every 1s
 */
// [[[$enclosing && $article < 9]]]
#define LOAD_DISPLAY_BASELINE 16
// [[[]]]
// [[[$enclosing && $article >= 9]]]
#define LOAD_DISPLAY_BASELINE  4
// [[[]]]
#define LOAD_DISPLAY_STARTROW 34
static void stat_thread(void * unused) __attribute__((noreturn));
static void stat_thread(void * unused)
{
  while (1)
    {
      sos_ui32_t flags;
      sos_ui32_t load1, load5, load15;
      char str1[11], str5[11], str15[11];
      struct sos_time t;
      t.sec = 1;
      t.nanosec = 0;

      sos_thread_sleep(& t);

      sos_disable_IRQs(flags);

      /* The IDLE task is EXcluded in the following computation */
      sos_load_get_sload(&load1, &load5, &load15);
      sos_load_to_string(str1, load1);
      sos_load_to_string(str5, load5);
      sos_load_to_string(str15, load15);
      sos_x86_videomem_printf(LOAD_DISPLAY_BASELINE+0, LOAD_DISPLAY_STARTROW,
			      SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			      "Kernel (- Idle): %s %s %s    ",
			      str1, str5, str15);

      sos_load_get_uload(&load1, &load5, &load15);
      sos_load_to_string(str1, load1);
      sos_load_to_string(str5, load5);
      sos_load_to_string(str15, load15);
      sos_x86_videomem_printf(LOAD_DISPLAY_BASELINE+1, LOAD_DISPLAY_STARTROW,
			      SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			      "User: %s %s %s        ",
			      str1, str5, str15);

      sos_load_get_uratio(&load1, &load5, &load15);
      sos_load_to_string(str1, load1);
      sos_load_to_string(str5, load5);
      sos_load_to_string(str15, load15);
      sos_x86_videomem_printf(LOAD_DISPLAY_BASELINE+2, LOAD_DISPLAY_STARTROW,
			      SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			      "User CPU %%: %s %s %s        ",
			      str1, str5, str15);

      /* The IDLE task is INcluded in the following computation */
      sos_load_get_sratio(&load1, &load5, &load15);
      sos_load_to_string(str1, load1);
      sos_load_to_string(str5, load5);
      sos_load_to_string(str15, load15);
      sos_x86_videomem_printf(LOAD_DISPLAY_BASELINE+3, LOAD_DISPLAY_STARTROW,
			      SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			      "Kernel CPU %% (+ Idle): %s %s %s    ",
			      str1, str5, str15);
      // [[[$enclosing && $master]]]
/*       extern void sos_process_dumplist(void); */
/*       sos_process_dumplist(); */
      // [[[]]]
      sos_restore_IRQs(flags);
    }
}

// [[[]]]
// [[[$article >= 7.5]]]

/* ======================================================================
 * Start the "init" (userland) process
 */
// [[[$enclosing && $article < 8]]]
static sos_ret_t start_init()
// [[[]]]
// [[[$enclosing && $article >= 8]]]
static sos_ret_t
start_init(struct sos_fs_manager_instance * rootfs)
// [[[]]]
{
  sos_ret_t retval;
  struct sos_umem_vmm_as *as_init;
  struct sos_process *proc_init;
  struct sos_thread *new_thr;
  sos_uaddr_t ustack, start_uaddr;
  int fake_args[32]; /* Must be large enough to contain argv/envp for init */
  sos_ret_t ret;
// [[[$enclosing && $article >= 8]]]
  struct sos_fs_opened_file * init_root, * init_cwd, * unused_of;
// [[[]]]

  /* Create the new process */
  proc_init = sos_process_create("init", FALSE);
  if (! proc_init)
    return -SOS_ENOMEM;
  as_init = sos_process_get_address_space(proc_init);

// [[[$enclosing && $article >= 8]]]

  /*
   * Setup the root and CWD directories of the process. The root of
   * this process will correspond to the "global" root of the whole
   * system since all the future processes will duplicate it !
   */
  retval = sos_fs_new_opened_file(proc_init, rootfs->root,
				  SOS_FS_OPEN_READ | SOS_FS_OPEN_WRITE,
				  & init_root);
  if (SOS_OK != retval)
    {
      sos_process_unref(proc_init);
      return -SOS_ENOENT;
    }

  /* Duplicate the root file to set the current working directory of
     the init process */
  retval = sos_fs_duplicate_opened_file(init_root, proc_init,
					& init_cwd);
  if (SOS_OK != retval)
    {
      sos_fs_close(init_root);
      sos_process_unref(proc_init);
      return -SOS_ENOENT;
    }

  /* Now update the process ! */
  if ( ( SOS_OK != sos_process_chroot(proc_init, init_root, & unused_of) )
       || ( SOS_OK != sos_process_chdir(proc_init, init_cwd, & unused_of) ) )
    {
      sos_fs_close(init_root);
      sos_fs_close(init_cwd);
      sos_process_chroot(proc_init, NULL, & unused_of);
      sos_process_chdir(proc_init, NULL, & unused_of);
      sos_process_unref(proc_init);
      return -SOS_ENOENT;
    }

// [[[]]]

  /* Map the 'init' program in user space */
  start_uaddr = sos_binfmt_elf32_map(as_init, "init");
  if (0 == start_uaddr)
    {
      sos_process_unref(proc_init);
      return -SOS_ENOENT;
    }

  /* Allocate the user stack */
  ustack = (SOS_PAGING_UPPER_USER_ADDRESS - SOS_DEFAULT_USER_STACK_SIZE) + 1;
  retval = sos_dev_zero_map(as_init, &ustack, SOS_DEFAULT_USER_STACK_SIZE,
			    SOS_VM_MAP_PROT_READ | SOS_VM_MAP_PROT_WRITE,
			    /* PRIVATE */ 0);
  if (SOS_OK != retval)
    {
      sos_process_unref(proc_init);
      return -SOS_ENOMEM;
    }

  /* Compute the address of the stack that will be used to initialize
     the user thread */
  ustack = SOS_ALIGN_INF((ustack + SOS_DEFAULT_USER_STACK_SIZE
			  - sizeof(fake_args)), 4);

  /* Build fake argv/envp arguments for the init process. See
     userland/crt.c for the format. */
  fake_args[0] = 0x1;                      /* argc */
  fake_args[1] = 4 * sizeof(fake_args[0]); /* offset of argv[0] */
  fake_args[2] = 0x0;                      /* delimiter between argv and envp */
  fake_args[3] = 0x0;                      /* end of envp */
  strzcpy ((char *) & fake_args[4],        /* argv[0] contents */
	   "init", 5);

  /* Copy the arguments to the user thread stack */
  ret = sos_memcpy_to_specified_userspace (as_init,
					   ustack,
					   (sos_vaddr_t) fake_args,
					   sizeof(fake_args));
  if (sizeof(fake_args) != ret)
    {
      sos_bochs_printf ("sos_memcpy_to_specified_userspace() failed, returned %d\n", ret);
      sos_process_unref(proc_init);
      return ret;
    }

  /* Now create the user thread */
  new_thr = sos_create_user_thread(NULL,
				   proc_init,
				   start_uaddr,
				   0, 0,
				   ustack,
				   SOS_SCHED_PRIO_TS_LOWEST);
  if (! new_thr)
    {
      sos_process_unref(proc_init);
      return -SOS_ENOMEM;
    }

  sos_process_unref(proc_init);
  return SOS_OK;
}

// [[[]]]

// [[[$enclosing && $article >= 11]]]

sos_ret_t socket_test (void)
{
  int sock;
  sos_ret_t ret;
  char *buf = "machin !";
  struct sockaddr_in sockaddr;
  ret = socket_create (PF_INET, SOCK_DGRAM, 0, & sock);
  if (SOS_OK != ret)
    {
      sos_bochs_printf ("Socket creation failed, %d\n", ret);
      return ret;
    }

  sockaddr.family = PF_INET;
  sockaddr.port   = 654;
  sockaddr.addr   = (172 << 24) | (20 << 16) | (1);

  ret = socket_sendto (sock, (sos_uaddr_t) buf, strlen(buf), & sockaddr,
		       sizeof (sockaddr));
  if (SOS_OK != ret)
    {
      sos_bochs_printf ("Socket send failed, %d\n", ret);
      return ret;
    }

  ret = socket_close (sock);
  if (SOS_OK != ret)
    {
      sos_bochs_printf ("Socket close failed, %d\n", ret);
      return ret;
    }

  return SOS_OK;
}

// [[[]]]

/* ======================================================================
 * The C entry point of our operating system
 */
void sos_main(unsigned long magic, unsigned long arg)
     __attribute__((noreturn));
void sos_main(unsigned long magic, unsigned long arg)
{
  /* [[[$article == 2]]] */ unsigned i;
// [[[$article >= 3]]]
  sos_paddr_t sos_kernel_core_base_paddr, sos_kernel_core_top_paddr;
// [[[]]]
// [[[$article == 6]]]
  struct syntax_node *syntax_tree;
// [[[]]]
// [[[$article >= 6.5]]]
  struct sos_time tick_resolution;
// [[[]]]
// [[[$article >= 8]]]
  struct sos_fs_manager_instance * rootfs;
// [[[]]]

  /* Size of RAM above 1MB. Might be undefined ! */
  unsigned long int upper_mem = 0;

  /* Setup bochs and console, and clear the console */
  sos_bochs_subsystem_setup();

  sos_x86_videomem_setup();
  sos_x86_videomem_cls(SOS_X86_VIDEO_BG_BLUE);
// [[[$master]]]

  sos_serial_subsystem_setup(TRUE);
  sos_serial_printf(0, "Bonjour\n");
// [[[]]]

  /* Greetings from SOS */
  if (magic == MULTIBOOT_BOOTLOADER_MAGIC)
    {
      /* Grub sends us a structure, called multiboot_info_t with a lot of
	 precious informations about the system, see the multiboot
	 documentation for more information. */
      multiboot_info_t *mbi = (multiboot_info_t *) arg;

      /* Multiboot says: "The value returned for upper memory is
	 maximally the address of the first upper memory hole minus 1
	 megabyte.". It also adds: "It is not guaranteed to be this
	 value." aka "YMMV" ;) */
      upper_mem = mbi->mem_upper;
      sos_x86_videomem_printf(1, 0,
			      SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			      "Welcome From GRUB to %s%c RAM is %dMB (upper mem = 0x%x kB)",
			      "SOS article @[$article]@@[" MASTER" if $master;]@", ',',
			      (unsigned)(upper_mem >> 10) + 1,
			      (unsigned)upper_mem);
    }
  else if (magic == 0x42244224)
    {
      /* Loaded with SOS bootsect */
      upper_mem = arg;
      sos_x86_videomem_printf(1, 0,
			      SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			      "Welcome to %s%c RAM is %dMB (upper mem = 0x%x kB)",
			      "SOS article @[$article]@@[" MASTER" if $master;]@", ',',
			      (unsigned)(upper_mem >> 10) + 1,
			      (unsigned)upper_mem);
    }
  else
    /* Not loaded with grub, not from an enhanced bootsect */
    sos_x86_videomem_printf(1, 0,
			    SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			    "Welcome to SOS article @[$article]@@[" MASTER" if $master;]@");

  sos_bochs_putstring("Message in a bochs: This is SOS article @[$article]@@[" MASTER" if $master;]@.\n");

// [[[$article >= 2]]]
  /* Setup CPU segmentation and IRQ subsystem */
  sos_gdt_subsystem_setup();
  sos_idt_subsystem_setup();

  /* Setup SOS IRQs and exceptions subsystem */
  sos_exception_subsystem_setup();
  sos_irq_subsystem_setup();

  /* Configure the timer so as to raise the IRQ0 at a 100Hz rate */
  sos_i8254_set_frequency(100);

//[[[]]]
// [[[$article >= 6.5]]]
  /* Setup the kernel time subsystem to get prepared to take the timer
     ticks into account */
  tick_resolution = (struct sos_time) { .sec=0, .nanosec=10000000UL };
  sos_time_subsysem_setup(& tick_resolution);

// [[[]]]
// [[[$article == 1]]]
  /* An operatig system never ends */
  for (;;)
    continue;
// [[[]]]
  //[[[$article >= 3]]]
  /* We need to know the RAM size */
  if (upper_mem == 0)
    {
      sos_x86_videomem_putstring(20, 0,
				 SOS_X86_VIDEO_FG_LTRED
				   | SOS_X86_VIDEO_BG_BLUE
				   | SOS_X86_VIDEO_FG_BLINKING,
				 "I don't know RAM size ! Load me with Grub...");
      /* STOP ! */
      for (;;)
	continue;
    }

  // [[[]]]
  // [[[$article >= 2]]]
  /*
   * Some interrupt handlers
   */

  /* Binding some HW interrupts and exceptions to software routines */
  sos_irq_set_routine(SOS_IRQ_TIMER,
		      clk_it);
  // [[[]]]
  // [[[$article == 2]]]
  sos_exception_set_routine(SOS_EXCEPT_DIVIDE_ERROR,
			    divide_ex);

  /*
   * Enabling the HW interrupts here, this will make the timer HW
   * interrupt call our clk_it handler
   */
  asm volatile ("sti\n");

  /* Raise a rafale of 'division by 0' exceptions. All this code is
     not really needed (equivalent to a bare "i=1/0;"), except when
     compiling with -O3: "i=1/0;" is considered dead code with gcc
     -O3. */
  i = 10;
  while (1)
    {
      /* Stupid function call to fool gcc optimizations */
      sos_bochs_printf("i = 1 / %d...\n", i);
      i = 1 / i;
    }

  /* Will never print this since the "divide by zero" exception always
     returns to the faulting instruction (see Intel x86 doc vol 3,
     section 5.12), thus re-evaluating the "divide-by-zero" exprssion
     and raising the "divide by zero" exception again and again... */
  sos_x86_videomem_putstring(2, 0,
			     SOS_X86_VIDEO_FG_LTRED | SOS_X86_VIDEO_BG_BLUE,
			     "Invisible");
  // [[[]]]
  // [[[$article >= 3]]]

  /*
   * Setup physical memory management
   */

  SOS_ASSERT_FATAL(SOS_OK
		   == sos_physmem_subsystem_setup((upper_mem<<10) + (1<<20),
						  &sos_kernel_core_base_paddr,
						  &sos_kernel_core_top_paddr));
  // [[[]]]
  // [[[$article == 3]]]
  test_physmem();
  // [[[]]]
  // [[[($article >= 4) && ($article < 6)]]]
  
  /*
   * Switch to paged-memory mode
   */

  /* Disabling interrupts should seem more correct, but it's not really
     necessary at this stage */
  if (sos_paging_subsystem_setup(sos_kernel_core_base_paddr,
				 sos_kernel_core_top_paddr))
    sos_bochs_printf("Could not setup paged memory mode\n");
  sos_x86_videomem_printf(2, 0,
			  SOS_X86_VIDEO_FG_YELLOW | SOS_X86_VIDEO_BG_BLUE,
			  "Paged-memory mode is activated");

  // [[[]]]
  // [[[$article >= 6]]]
  
  /*
   * Switch to paged-memory mode
   */

  /* Disabling interrupts should seem more correct, but it's not really
     necessary at this stage */
  SOS_ASSERT_FATAL(SOS_OK ==
		   sos_paging_subsystem_setup(sos_kernel_core_base_paddr,
					      sos_kernel_core_top_paddr));
  
  // [[[]]]
  // [[[$article == 4]]]
  test_paging(sos_kernel_core_top_paddr);
  // [[[]]]
  // [[[$article >= 6 ]]]
  /* Bind the page fault exception */
  sos_exception_set_routine(SOS_EXCEPT_PAGE_FAULT,
			    pgflt_ex);
  // [[[]]]
  // [[[$article >= 5 ]]]

  /*
   * Setup kernel virtual memory allocator
   */

  if (sos_kmem_vmm_subsystem_setup(sos_kernel_core_base_paddr,
				   sos_kernel_core_top_paddr,
				   bootstrap_stack_bottom,
				   bootstrap_stack_bottom
				   + bootstrap_stack_size))
    sos_bochs_printf("Could not setup the Kernel virtual space allocator\n");

  if (sos_kmalloc_subsystem_setup())
    sos_bochs_printf("Could not setup the Kmalloc subsystem\n");
  // [[[]]]
  // [[[$article == 5]]]

  /* Run some kmalloc tests */
  bn_test();
  // [[[]]]
  // [[[$article >= 7 ]]]

  /*
   * Initialize the MMU context subsystem
   */
  sos_mm_context_subsystem_setup();

  /*
   * Initialize the CPU context subsystem
   */
  sos_cpu_context_subsystem_setup();

  /*
   * Bind the syscall handler to its software interrupt handler
   */
  sos_swintr_subsystem_setup();

  // [[[]]]
  // [[[$article >= 6.5 ]]]

  /*
   * Initialize the Kernel thread and scheduler subsystems
   */
  
  /* Initialize kernel thread subsystem */
  sos_thread_subsystem_setup(bootstrap_stack_bottom,
			     bootstrap_stack_size);

  /* Initialize the scheduler */
  sos_sched_subsystem_setup();

  /* Declare the IDLE thread */
  // [[[$enclosing && $article <= 6.5]]]
  SOS_ASSERT_FATAL(sos_create_kernel_thread("idle", idle_thread, NULL) != NULL);
  // [[[]]]
  // [[[$enclosing && $article > 6.5]]]
  SOS_ASSERT_FATAL(sos_create_kernel_thread("idle", idle_thread, NULL,
   					    SOS_SCHED_PRIO_TS_LOWEST) != NULL);
  // [[[]]]
  // [[[]]]
  // [[[$article >= 7]]]

  /* Prepare the stats subsystem */
  sos_load_subsystem_setup();

  /* Declare a thread that prints some stats */
  SOS_ASSERT_FATAL(sos_create_kernel_thread("stat_thread", stat_thread,
					    NULL,
					    SOS_SCHED_PRIO_TS_LOWEST) != NULL);

  // [[[$enclosing && $article >= 7.5]]]

  /*
   * Initialise user address space management subsystem
   */
  sos_umem_vmm_subsystem_setup();
  sos_dev_zero_subsystem_setup();
  // [[[]]]
  // [[[$article >= 9.5]]]
  
  /* Initialize the page and block cache subsystems */
  sos_fs_pagecache_subsystem_setup();
  sos_blockdev_subsystem_setup();
  // [[[]]]
  // [[[$article >= 9]]]
  sos_dev_mem_chardev_setup();
  // [[[]]]

  /*
   * Initialize process stuff
   */
  sos_process_subsystem_setup();

  // [[[]]]
  // [[[$article > 2 && $article < 6.5]]]

  /*
   * Enabling the HW interrupts here, this will make the timer HW
   * interrupt call our clk_it handler
   */
  asm volatile ("sti\n");
  // [[[]]]
  // [[[$article == 6]]]

  /*
   * Print hello world using coroutines
   */
  print_hello_world();


  /*
   * Run coroutine tests
   */
  sos_x86_videomem_printf(4, 0,
			  SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_LTGREEN,
			  "Coroutine test");
  sos_x86_videomem_printf(5, 0,
			  SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_YELLOW,
			  "Parsing...");
  syntax_tree = parse_expression(" -  ( (69/ toto)+ ( (( - +-- 1))) + --toto*((toto+ - - y - +2*(y-toto))*y) +2*(y-toto) )/- (( y - toto)*2)");

  if (syntax_tree != NULL)
    {
      sos_x86_videomem_printf(6, 0,
			      SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_YELLOW,
			      "Evaluating...");
      sos_x86_videomem_printf(7, 0,
			      SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_YELLOW,
			      "Result=%d (if 0: check bochs output)",
			      eval_expression(syntax_tree,
					      (const char*[]){"toto", "y"},
					      (int[]){3, 4},
					      2));
      free_syntax_tree(syntax_tree);
      sos_x86_videomem_printf(8, 0,
			      SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_YELLOW,
			      "Done (un-allocated syntax tree)");
    }
  else
    {
      sos_x86_videomem_printf(6, 0,
			      SOS_X86_VIDEO_BG_BLUE | SOS_X86_VIDEO_FG_YELLOW,
			      "Error in parsing (see bochs output)");
    }

  /*
   * Run some demand-paging tests
   */
  test_demand_paging(234567, 500);


  /*
   * Create an un-resolved page fault, which will make the page fault
   * handler print the backtrace.
   */
  test_backtrace(6, 0xdeadbeef, bootstrap_stack_bottom, bootstrap_stack_size);

  /*
   * System should be halted BEFORE here !
   */

  // [[[]]]
  // [[[$article == 6 && $master]]]
  extern void test_art6();
  // test_art6();
  // [[[]]]
  // [[[$article >= 6.5 ]]]

  /* Enabling the HW interrupts here, this will make the timer HW
     interrupt call the scheduler */
  asm volatile ("sti\n");

  // [[[]]]

  // [[[$article >= 8]]]
  SOS_ASSERT_FATAL(SOS_OK == sos_fs_virtfs_subsystem_setup());
  SOS_ASSERT_FATAL(SOS_OK == sos_fs_subsystem_setup(NULL,
						    "virtfs",
						    NULL,
						    & rootfs));

  // [[[]]]

  // [[[$article >= 9]]]
  tty_subsystem_setup();
  sos_ttyS0_subsystem_setup();
  sos_console_subsystem_setup();

  // [[[]]]

  // [[[$article >= 9.5]]]
  sos_ide_subsystem_setup();

  // [[[]]]

  // [[[$article >= 10]]]
  SOS_ASSERT_FATAL(SOS_OK == sos_fs_ext2fs_subsystem_setup());
  // [[[]]]

  // [[[$article >= 11]]]

  sos_pci_subsystem_setup();
  sos_ne2k_pci_init("eth0", ((172 << 24) | (20 << 16) | (2)));
  socket_test();

  // [[[]]]

  // [[[$article == 7.5]]]
  /* Start the 'init' process, which in turns launches the other
     programs */
  start_init();
  // [[[]]]
  // [[[$article >= 8]]]
  /* Start the 'init' process, which in turns launches the other
     programs */
  start_init(rootfs);
  // [[[]]]
  // [[[$article >= 7 && $article < 7.5]]]
  /* Run some tests involving USER processes and threads */
  extern void test_art7(void);
  test_art7();
  // [[[]]]
  // [[[($article >= 6.5) && ($master || $article < 7.5)]]]

  /* Now run some Kernel threads just for fun ! */
  extern void MouseSim(void);
  // MouseSim();
  //  test_thread();
  // [[[]]]
  // [[[$article >= 6.5 && $master]]]
  extern void test_art6_5(void);
  // test_art6_5();
  // [[[]]]
  // [[[$article >= 5 && $master]]]
  extern void test_art5(void);
  // test_art5();
  //  test_art5();
  //  test_art5();
  //  test_art5();
  //  test_art5();
  // [[[]]]
  // [[[$enclosing && $article > 1 && $article < 6.5]]]

  /* An operatig system never ends */
  for (;;)
    {
      /* Remove this instruction if you get an "Invalid opcode" CPU
	 exception (old 80386 CPU) */
      asm("hlt\n");

      continue;
    }
  // [[[]]]
  // [[[$enclosing && $article >= 6.5]]]
  /*
   * We can safely exit from this function now, for there is already
   * an idle Kernel thread ready to make the CPU busy working...
   *
   * However, we must EXPLICITELY call sos_thread_exit() because a
   * simple "return" will return nowhere ! Actually this first thread
   * was initialized by the Grub bootstrap stage, at a time when the
   * word "thread" did not exist. This means that the stack was not
   * setup in order for a return here to call sos_thread_exit()
   * automagically. Hence we must call it manually. This is the ONLY
   * kernel thread where we must do this manually.
   */
  sos_bochs_printf("Bye from primary thread !\n");
  sos_thread_exit();
  SOS_FATAL_ERROR("No trespassing !");
  // [[[]]]
}
