// [[[$article >= 11]]]
/* Copyright (C) 2005  Thomas Petazzoni

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/

#include <sos/types.h>
#include <sos/socket.h>
#include <sos/kmalloc.h>
#include <drivers/netdev_generic.h>
#include <drivers/net_protos.h>

struct socket {
  int domain;
  int type;
  int port;
};

#define MAX_SOCKETS 16
struct socket *sockets[MAX_SOCKETS];

sos_ret_t socket_create (int domain, int type, int protocol, int *fd)
{
  struct socket *sk;
  int i;

  if (domain != PF_INET)
    return -SOS_EINVAL;

  if (type != SOCK_DGRAM)
    return -SOS_EINVAL;

  sk = (struct socket *) sos_kmalloc (sizeof(struct socket), 0);
  if (! sk)
    return -SOS_ENOMEM;

  sk->domain = domain;
  sk->type   = type;
  sk->port   = 0;

  for (i = 0; i < MAX_SOCKETS; i++)
    {
      if (! sockets[i])
	{
	  sockets[i] = sk;
	  *fd = i;
	  return SOS_OK;
	}
    }

  sos_kfree ((sos_vaddr_t) sk);

  return -SOS_ENOMEM;
}

sos_ret_t socket_close (int fd)
{
  if (fd >= MAX_SOCKETS || fd < 0)
    return -SOS_EINVAL;

  if (sockets[fd] == NULL)
    return -SOS_EBADF;

  sos_kfree ((sos_vaddr_t) sockets[fd]);
  sockets[fd] = NULL;

  return SOS_OK;
}

sos_ret_t socket_sendto (int fd, sos_uaddr_t buf, sos_size_t len,
			 const struct sockaddr_in *to, sos_size_t tolen)
{
  struct sos_netdev *dev;
  struct sos_netdev_pkt *pkt;

  if (fd >= MAX_SOCKETS || fd < 0)
    return -SOS_EINVAL;

  if (sockets[fd] == NULL)
    return -SOS_EBADF;

  dev = sos_netdev_default();
  if (! dev)
    return -SOS_EBADF;

  pkt = sos_udp_alloc_pkt (dev, len, TRUE, 0);
  if (! pkt)
    return -SOS_ENOMEM;

  /* FIXME: copy_from_user */
  memcpy (& pkt->contents[pkt->pdu_offset],
	  (void *) buf, len);

  /* Packet will be freed by the TX thread */
  return sos_udp_send_packet (dev, pkt,
			      5421, to->port,
			      NETADDR_2_IPV4(dev->netaddr), to->addr);
}

sos_ret_t socket_bind (int fd, const struct sockaddr_in *my_addr,
		       sos_size_t addrlen)
{
  if (fd >= MAX_SOCKETS || fd < 0)
    return -SOS_EINVAL;

  if (sockets[fd] == NULL)
    return -SOS_EBADF;

  sockets[fd]->port = ntohs(my_addr->port);

  return SOS_OK;
}

sos_ret_t socket_recvfrom (int fd, sos_uaddr_t buf, sos_size_t len,
			   struct sockaddr *from, sos_size_t fromlen)
{
  if (fd >= MAX_SOCKETS || fd < 0)
    return -SOS_EINVAL;

  if (sockets[fd] == NULL)
    return -SOS_EBADF;

}

// [[[]]]
