// [[[$article >= 6.5 && $master]]]


#include <drivers/bochs.h>
#include <sos/assert.h>

#include <hwcore/irq.h>
#include <drivers/x86_videomem.h>

#include <sos/time.h>
#include <sos/thread.h>
#include <sos/ksynch.h>

extern int rnd_int1(int max); /* test-art5.c */

static void timeout1(struct sos_timeout_action *act)
{
  struct sos_time now, delay;
  sos_time_get_now(& now);

  sos_bochs_printf("Bonjour %x (now=%ds %dns) data=%d\n",
		   (unsigned)act, (int)now.sec, (int)now.nanosec,
		   (int)act->routine_data);

  /* Re-program timeout */
  delay = (struct sos_time) { .sec=3, .nanosec=0 };
  //  delay = (struct sos_time) { .sec=0, .nanosec=1000000 };
  //  sos_time_get_tick_resolution(& delay);
  sos_time_register_action_relative(act, & delay, timeout1, act->routine_data);
}


static void print_time(const char * s, const struct sos_time *t)
     __attribute__ ((unused));

static void print_time(const char * s, const struct sos_time *t)
{
  sos_bochs_printf("TIME[%s]: %ds %dns\n", s, (int)t->sec, (int)t->nanosec);
}


static struct sos_timeout_action act1, act2, act3, act4;


static void test_time(void)
{
  struct sos_time delay;

  {
    struct sos_time orig_t1, t1, t2;

    orig_t1 = t1 = (struct sos_time) { .sec=3, .nanosec=21 };
    t2 = (struct sos_time) { .sec=1, .nanosec=42 };
    SOS_ASSERT_FATAL(sos_time_cmp(& t1, & t2) > 0);
    SOS_ASSERT_FATAL(sos_time_cmp(& t2, & t1) < 0);
    SOS_ASSERT_FATAL(sos_time_cmp(& t1, & t1) == 0);
    SOS_ASSERT_FATAL(sos_time_cmp(& t2, & t2) == 0);

    sos_time_dec(& t1, & t2);
    SOS_ASSERT_FATAL(sos_time_cmp(& t1, & orig_t1) < 0);

    sos_time_inc(& t1, & t2);
    SOS_ASSERT_FATAL(sos_time_cmp(& t1, & orig_t1) == 0);
  }

  delay.sec = 2;
  delay.nanosec = 20000000;
  sos_time_init_action(& act4);
  sos_time_register_action_relative(& act4, & delay, timeout1, (void*)4);

  delay.sec = 2;
  delay.nanosec = 0;
  sos_time_init_action(& act2);
  sos_time_register_action_relative(& act2, & delay, timeout1, (void*)2);

  delay.sec = 2;
  delay.nanosec = 0;
  sos_time_init_action(& act3);
  sos_time_register_action_relative(& act3, & delay, timeout1, (void*)3);

  delay.sec = 1;
  delay.nanosec = 0;
  sos_time_init_action(& act1);
  sos_time_register_action_relative(& act1, & delay, timeout1, (void*)1);

  sos_bochs_printf("GO !\n");
}


#define sos_bochs_printf(fmt,args...) \
  (sos_bochs_printf)("THR %p: " fmt, sos_thread_get_current(), ##args)


#define NB_THREADS 200
static struct sos_ksema  sem_exit_main;
static struct sos_ksema  sem_rdv;
static struct sos_kmutex lck_rdv;
static int               nb_waiting_rdv;

static void rendez_vous(void)
{
  sos_bool_t is_last;

  SOS_ASSERT_FATAL(SOS_OK == sos_kmutex_lock(& lck_rdv, NULL));
  nb_waiting_rdv ++;
  is_last = (nb_waiting_rdv == NB_THREADS);
  SOS_ASSERT_FATAL(SOS_OK == sos_kmutex_unlock(& lck_rdv));

  if (is_last)
    {
      /* Wakeup the other NB_THREADS threads */
      int i;
      for (i = 0 ; i < NB_THREADS - 1 ; i++)
	SOS_ASSERT_FATAL(SOS_OK == sos_ksema_up(& sem_rdv));
    }
  else if (rnd_int1(2) == 0)
    {
      struct sos_time t = (struct sos_time) { .sec=0, .nanosec=10000000 };
      /* Wait for the last one to wake us up */
      while (SOS_OK != sos_ksema_down(& sem_rdv, &t))
	{
	  SOS_ASSERT_FATAL(sos_time_is_zero(& t));
	  (sos_bochs_printf)("T");
	  t = (struct sos_time) { .sec=0, .nanosec=30000000 };
	}
    }
  else
    SOS_ASSERT_FATAL(SOS_OK == sos_ksema_down(& sem_rdv, NULL));
}


static void thr_func(void *unused)
{
  int i;
  sos_bochs_printf("Thr: Fait des choses avant le rendez-vous...\n");
  for (i = 0 ; i < NB_THREADS+42 ; i++)
    if (rnd_int1(2) == 0)
      {
	static int yo;
	struct sos_time t = { .sec=0, .nanosec=20000000 };
	SOS_ASSERT_FATAL(SOS_OK == sos_thread_sleep(& t));
	SOS_ASSERT_FATAL(sos_time_is_zero(& t));
	yo = (yo + 1) %2;
	(sos_bochs_printf)("%c", yo?'-':'+');
      }
    else
      sos_thread_yield();

  sos_bochs_printf("---------- Thr: Attente des autres threads...\n");
  rendez_vous();
  sos_bochs_printf("++++++++++ Thr: Ok : attente sur rdv terminee\n");

  sos_bochs_printf("Thr: Fait des choses apres le rendez-vous...\n");
  for (i = 0 ; i < NB_THREADS+42 ; i++)
    if (rnd_int1(2) == 0)
      {
	static int yo;
	struct sos_time t = { .sec=0, .nanosec=20000000 };
	SOS_ASSERT_FATAL(SOS_OK == sos_thread_sleep(& t));
	SOS_ASSERT_FATAL(sos_time_is_zero(& t));
	yo = (yo + 1) %2;
	(sos_bochs_printf)("%c", yo?'-':'+');
      }
    else
      sos_thread_yield();

  sos_bochs_printf("Thr: Indique au main qu'on a termine\n");
  sos_ksema_up(& sem_exit_main);
}


static void test_synchro(void)
{
  int i;
#define MY_KWQ_ORDER SOS_KWQ_ORDER_PRIO

  sos_bochs_printf("** ** ** Bonjour le monde ** ** **\n");

  /* Creation du semaphore de sortie du main */
  sos_ksema_init(& sem_exit_main, "exit_main", 0, MY_KWQ_ORDER);

  nb_waiting_rdv = 0;
  sos_ksema_init(& sem_rdv, "sem_rdv", 0, MY_KWQ_ORDER);
  sos_kmutex_init(& lck_rdv, "lck_rdv", MY_KWQ_ORDER);
  sos_bochs_printf("main: sema/mutex Ok\n");

  /* Creation des threads de demo */
  for (i = 0 ; i < NB_THREADS ; i++)
    {
      struct sos_thread *thr;
      thr = sos_create_kernel_thread("a thread", thr_func,
				     NULL, SOS_SCHED_PRIO_TS_LOWEST);
      sos_bochs_printf("main: Nouveau thread %d = 0x%x\n", i, (unsigned)thr);

      // Pour s'amuser a voir les fils entrelacer la creation des autres fils
      sos_thread_yield();
    }

  /* Attente de la fin des fils */
  sos_bochs_printf("main: Attente de la fin des fils...\n");
  for (i = 0 ; i < NB_THREADS ; i++)
    {
      sos_ksema_down(& sem_exit_main, NULL);
    }

  sos_kmutex_dispose(& lck_rdv);
  sos_ksema_dispose(& sem_rdv);
  sos_ksema_dispose(& sem_exit_main);

  sos_bochs_printf("** ** ** Au revoir le monde ** ** **\n");
}


void test_art6_5(void)
{
  test_time();
  test_synchro();
}

// [[[]]]
