# [[[0]]]

## Script to extract portions of a file with special tags containing
## tests. The tags form *nested blocks*. The special variable
## "$enclosing" holds the condition value for the upper nesting level.
##
## Conditional block discard:
##   // [[[perl condition]]]
##   ...
##   // [[[]]]
##
## 1 liner conditional discard:
##   /* [[[perl condition]]] */ 1liner
##
## Substitution:
##   "Yo @[1+2]@ Yo"
##   ===> "Yo 3 Yo"
##
## example:
##   shell> ./extract.pl '$version = 12'
##   /* [[[ 1<=2 ]]] */ a=b
##   // [[[ $version >= 3 ]]]
##   Mouf
##   // [[[ $enclosing && 2 >= 1 ]]]
##   Hugh
##   // [[[]]]
##   // [[[ $version < 10 ]]]
##   /* [[[$enclosing && 1<=2]]] */ b=c
##   Hello
##   // [[[ $enclosing && 2 >= 1 ]]]
##   Yo
##   // [[[]]]
##   // [[[]]]
##   // [[[]]]
## Will result in :
##   a=b
##   Mouf
##   Hugh
## Note : it will be Ok for '#' instead of '//'

# hash structure of boolean elements
# Condition for top stack is always true
%condstack;

%condstr;

$top_cond;

sub begin_block
{
    my ($cond,$index) = @_;

    # Prepare "$enclosing" condition variable
    my $enclosing = 1;
    $enclosing = $condstack{$top_cond} if ($top_cond > 0);

    # Add a new condition on the stack
    $top_cond ++;
    $condstack{$top_cond} = eval($cond);
    $condstr{$top_cond}   = "Condition \"" . $cond
	. "\" (near " . $orig_fname . ":" . $index . ")";
###    print "/// # BEGIN BLOCK " . $condstr{$top_cond} . "\n";
}

sub end_block
{
    my ($cond,$index) = @_;
    die "Problme d'imbrication des blocs, ligne $index\n" if ($top_cond < 1);

###    print "/// # END BLOCK " . $condstr{$top_cond} . "\n";
    $top_cond --;
}


sub end_parsefile
{
    die "Problme d'imbrication des blocs en fin de fichier\n"
	if ($top_cond < 0);
    if ($top_cond > 0)
    {
	print STDERR "Problme: blocs non termines !\n";
	for (my $i = 1 ; $i <= $top_cond ; $i++)
	{
	    print STDERR "  $i: $condstr{$i}\n";
	}
	die "Stop\n";
    }
}

sub keep_str
{
    my $str = shift;

    if ($condstack{$top_cond})
    {
	if (my @lstmatch = ($str =~ /[^@]*@\[([^@]+)\]@[^@]*/gio))
	{
	    foreach my $m (@lstmatch)
	    {
		my $evalm = eval($m);
		$str =~ s/@\[\Q$m\E\]@/$evalm/gi;
	    }
	}
	print "$str";
    }
}

sub parse_line
{
    my ($line, $index) = @_;
    my $pattarr;

    if (@pattarr = ($line =~ /^\s*(\/\/|\#)\s*\[\[\[(.*)\]\]\]\s*$/io) )
    {
	my ($comment, $cond) = @pattarr;
	# print "orig: " . $line;
	print "# $index \"$orig_fname\"\n" if (! ($orig_fname eq "-"));
	if ($cond eq "") { &end_block($cond, $index); }
	else  { &begin_block($cond, $index); }
    }

    elsif (@pattarr =
	   ($line =~ /^(\s*)\/\*\s*\[\[\[(.*)\]\]\]\s*\*\/\s*(.*)$/io) )
    {
	my ($space, $cond, $code) = @pattarr;
	# print "///////// 1line orig: $line";
	# print "///////// sp=-$sp-, cond=-$cond-, code=-$code-\n";
	my $enclosing = $condstack{$top_cond};
	if (eval($cond))
	{
	    print $space . $code . "\n";
	}
	else
	{
	    print "# $index \"$orig_fname\"\n" if (! ($orig_fname eq "-"));
	}
    }

    # Catch-up with the CPP location marks (# nb_line "file_name" pos?)
    elsif (@pattarr = ($line =~ /^\#\s+(\d+)\s+"([^]]+)"(\s+\d+)?$/) )
    {
	my ($i,$f) = @pattarr;
	# print "Diese <$i> -$f- (orig=$line)\n";
	$index = $i;
	$orig_fname = $f;

	&keep_str($line);
    }

    else
    {
	&keep_str($line);
    }

    return $index;
}

sub parse_stdin
{
    my $opts = shift;
    eval $opts;

    $top_cond             = 0;
    $condstack{$top_cond} = 1;
    $condstr{$top_cond}   = "<TOP>";

    my $line, $i=0;
    $curfile = "STDIN";
    while ($line = <STDIN>)
    {
	$i++;
	$i = &parse_line($line, $i);
    }
    end_parsefile;
}

$orig_fname = shift @ARGV;

foreach $arg (@ARGV)
{
    eval $arg;
}

&parse_stdin ();
# [[[]]]
