// [[[$article >= 7.5]]]
/* Copyright (C) 2005 David Decotigny

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA. 
*/

#include <libc.h>
#include <debug.h>
#include <string.h>
// [[[$enclosing && $article >= 9]]]
#include <drivers/devices.h>
#include "fstest_utils.h"
// [[[]]]

/**
 * @file init.c
 * Test fork and exec()
 */

int main(void)
{
  bochs_printf("init: Welcome in userland !\n");

// [[[$enclosing && $article >= 9]]]

  /* Creating initial device nodes */
  TEST_EXPECT_CONDITION(mkdir("/dev", S_IRWXALL), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/zero", S_IRUSR | S_IWUSR,
			      S_IFCHR, SOS_CHARDEV_ZERO_MAJOR, SOS_CHARDEV_ZERO_MINOR), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/null", S_IRUSR | S_IWUSR,
			      S_IFCHR, SOS_CHARDEV_ZERO_MAJOR, SOS_CHARDEV_NULL_MINOR), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/kmem", S_IRUSR | S_IWUSR,
			      S_IFCHR, SOS_CHARDEV_MEM_MAJOR, SOS_CHARDEV_KMEM_MINOR), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/mem", S_IRUSR | S_IWUSR,
			      S_IFCHR, SOS_CHARDEV_MEM_MAJOR, SOS_CHARDEV_PHYSMEM_MINOR), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/tty", S_IRUSR | S_IWUSR,
			      S_IFCHR, SOS_CHARDEV_TTY_MAJOR, SOS_CHARDEV_CONSOLE_MINOR), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/ttyS", S_IRUSR | S_IWUSR,
			      S_IFCHR, SOS_CHARDEV_TTY_MAJOR, SOS_CHARDEV_SERIAL_MINOR), RETVAL == 0);

// [[[$enclosing && $article >= 9.5]]]
  TEST_EXPECT_CONDITION(mknod("/dev/hda", S_IRUSR | S_IWUSR,
			      S_IFBLK, SOS_BLOCKDEV_IDE_MAJOR, 0), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/hda1", S_IRUSR | S_IWUSR,
			      S_IFBLK, SOS_BLOCKDEV_IDE_MAJOR, 1), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/hda2", S_IRUSR | S_IWUSR,
			      S_IFBLK, SOS_BLOCKDEV_IDE_MAJOR, 2), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/hda3", S_IRUSR | S_IWUSR,
			      S_IFBLK, SOS_BLOCKDEV_IDE_MAJOR, 3), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/hda4", S_IRUSR | S_IWUSR,
			      S_IFBLK, SOS_BLOCKDEV_IDE_MAJOR, 4), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/hda5", S_IRUSR | S_IWUSR,
			      S_IFBLK, SOS_BLOCKDEV_IDE_MAJOR, 5), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/hda6", S_IRUSR | S_IWUSR,
			      S_IFBLK, SOS_BLOCKDEV_IDE_MAJOR, 6), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/hda7", S_IRUSR | S_IWUSR,
			      S_IFBLK, SOS_BLOCKDEV_IDE_MAJOR, 7), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/hdb", S_IRUSR | S_IWUSR,
			      S_IFBLK, SOS_BLOCKDEV_IDE_MAJOR, 16), RETVAL == 0);
  TEST_EXPECT_CONDITION(mknod("/dev/hdb1", S_IRUSR | S_IWUSR,
			      S_IFBLK, SOS_BLOCKDEV_IDE_MAJOR, 17), RETVAL == 0);

// [[[]]]
  ls("/", 1, 1);

  /* Set up the shell on the console */
  TEST_EXPECT_CONDITION(open("/dev/tty", O_RDWR), RETVAL == 0);
  TEST_EXPECT_CONDITION(open("/dev/tty", O_RDWR), RETVAL == 1);
  TEST_EXPECT_CONDITION(open("/dev/tty", O_RDWR), RETVAL == 2);

  if (fork() == 0)
    exec ("shell");

  close (2);
  close (1);
  close (0);

  /* Set up the shell on the serial port */
  TEST_EXPECT_CONDITION(open("/dev/ttyS", O_RDWR), RETVAL == 0);
  TEST_EXPECT_CONDITION(open("/dev/ttyS", O_RDWR), RETVAL == 1);
  TEST_EXPECT_CONDITION(open("/dev/ttyS", O_RDWR), RETVAL == 2);

  if (fork() == 0)
    exec ("shell");

  close (2);
  close (1);
  close (0);

// [[[]]]

// [[[$enclosing && $article == 8]]]
  /* launch the tests */
  if (fork() == 0)
    exec ("fstest");

// [[[]]]
// [[[$enclosing && $article < 8]]]
  if (fork() == 0)
    exec("myprog7");

  if (fork() == 0)
    exec("myprog8");

  if (fork() == 0)
    exec("myprog9");

  if (fork() == 0)
    exec("myprog10");

  if (fork() == 0)
    exec("myprog11");

  if (fork() == 0)
    exec("myprog12");
 
  if (fork() == 0)
    exec("myprog13");

  if (fork() == 0)
    exec("myprog14");

// [[[]]]
// [[[$enclosing && $article != 8 && $article < 9]]]
  if (fork() == 0)
    exec("banner");

// [[[]]]

// [[[$enclosing && $article >= 10]]]
  {
    int ret, fd;
    char buffer1[16];
    char buffer2[16];
    char *ptr;

    /* Mount a partition */
    TEST_EXPECT_CONDITION(mkdir("/mnt", S_IRWXALL), RETVAL == 0);
    TEST_EXPECT_CONDITION(mount("/dev/hda", "/mnt", "ext2fs", 0, NULL), RETVAL == 0);

    /* Open a file, and map it */
    TEST_EXPECT_CONDITION(fd = open("/mnt/small", O_RDWR), RETVAL >= 0);
    TEST_EXPECT_CONDITION(ptr = mmap(NULL, 4096, PROT_READ | PROT_WRITE,
				     MAP_SHARED, fd, 0), RETVAL >= 0);

    /* Fetch file contents from memory */
    strzcpy(buffer1, ptr, sizeof(buffer1));

    /* Fetch file contents from file */
    TEST_EXPECT_CONDITION(ret = read(fd, buffer2, sizeof(buffer2)), RETVAL == sizeof(buffer2));
    buffer2[ret-1] = 0;

    /* Make sure both are coherent */
    TEST_EXPECT_CONDITION(strncmp(buffer1, buffer2, sizeof(buffer1)), RETVAL == 0);

    /* Write something new almost at the beginning of the file */
    TEST_EXPECT_CONDITION(lseek(fd, 2, SEEK_SET), RETVAL == 2);
    TEST_EXPECT_CONDITION(write(fd, "tatameuh", 8), RETVAL == 8);
    TEST_EXPECT_CONDITION(lseek(fd, 0, SEEK_SET), RETVAL == 0);

    /* Read again from the file, and check that the file correctly changed */
    TEST_EXPECT_CONDITION(ret = read(fd, buffer2, sizeof(buffer2)), RETVAL >= 0);
    buffer2[ret-1] = 0;
    TEST_EXPECT_CONDITION(strncmp(buffer2, buffer1, 2), RETVAL == 0);
    TEST_EXPECT_CONDITION(strncmp(buffer2 + 2, "tatameuh", 8), RETVAL == 0);
    TEST_EXPECT_CONDITION(strncmp(buffer2 + 10, buffer1 + 10, 6), RETVAL == 0);

    /* Read again from the memory, do the same checks */
    strzcpy(buffer2, ptr, sizeof(buffer2));
    TEST_EXPECT_CONDITION(strncmp(buffer2, buffer1, 2), RETVAL == 0);
    TEST_EXPECT_CONDITION(strncmp(buffer2 + 2, "tatameuh", 8), RETVAL == 0);
    TEST_EXPECT_CONDITION(strncmp(buffer2 + 10, buffer1 + 10, 6), RETVAL == 0);

    TEST_EXPECT_CONDITION(munmap(ptr, 4096), RETVAL == 0);
    TEST_EXPECT_CONDITION(close(fd), RETVAL == 0);
    TEST_EXPECT_CONDITION(umount("/mnt"), RETVAL == 0);

    /* Remount again, and do the same check once again */
    TEST_EXPECT_CONDITION(mount("/dev/hda", "/mnt", "ext2fs", 0, NULL), RETVAL == 0);

    /* Open a file, and map it */
    TEST_EXPECT_CONDITION(fd = open("/mnt/small", O_RDWR), RETVAL >= 0);
    TEST_EXPECT_CONDITION(ptr = mmap(NULL, 4096, PROT_READ | PROT_WRITE,
				     MAP_SHARED, fd, 0), RETVAL >= 0);

    /* Read from the file, and check that everything is still in
       place */
    TEST_EXPECT_CONDITION(ret = read(fd, buffer2, sizeof(buffer2)), RETVAL >= 0);
    buffer2[ret-1] = 0;
    TEST_EXPECT_CONDITION(strncmp(buffer2, buffer1, 2), RETVAL == 0);
    TEST_EXPECT_CONDITION(strncmp(buffer2 + 2, "tatameuh", 8), RETVAL == 0);
    TEST_EXPECT_CONDITION(strncmp(buffer2 + 10, buffer1 + 10, 6), RETVAL == 0);

    /* Read from memory, and do the same checks */
    strzcpy(buffer2, ptr, sizeof(buffer2));
    TEST_EXPECT_CONDITION(strncmp(buffer2, buffer1, 2), RETVAL == 0);
    TEST_EXPECT_CONDITION(strncmp(buffer2 + 2, "tatameuh", 8), RETVAL == 0);
    TEST_EXPECT_CONDITION(strncmp(buffer2 + 10, buffer1 + 10, 6), RETVAL == 0);

    TEST_EXPECT_CONDITION(munmap(ptr, 4096), RETVAL == 0);
    TEST_EXPECT_CONDITION(close(fd), RETVAL == 0);
    TEST_EXPECT_CONDITION(umount("/mnt"), RETVAL == 0);
  }
// [[[]]]

  bochs_printf("init: The end\n");
  return 0;
}
// [[[]]]
